local version = "1.5"
local cpath = "/www/server/total/"
if not package.cpath:find(cpath) then
	package.cpath = cpath .. "?.so;" .. package.cpath ..";;"
end
if not package.path:find(cpath) then
	package.path = cpath .. "?.lua;" .. package.path
end
local server_name,ip,today,day,body_length,config,method,httpd,cache,tms,cache_count
local cache_count_threshold, cache_interval_threshold = 300, 10
local cjson, sqlite3, memcached, socket
local total_config
local update_day
local data_dir
local no_binding_server_name 
local number_day
local day_column
local flow_column
local spider_column

local debug_mode = false

local function debug(msg)
	if not debug_mode then return true end
    local fp = io.open('/www/server/total/debug.log', 'ab')
    if fp == nil then
         return nil
    end
	local localtime = os.date("%Y-%m-%d %H:%M:%S")
	if server_name then
    	fp:write(server_name.."/"..localtime..":"..tostring(msg) .. "\n")
	else
    	fp:write(localtime..":"..tostring(msg) .. "\n")
	end
    fp:flush()
    fp:close()
    return true
end

local function file_exists(file)
	local file = io.open(file, "rb")
	if file then 
		file:close()
		return true 
	end
	return false
end

function check_dir(path)
	local file = io.open(path, "rb")
	if file then file:close() end
	return file ~= nil
end

function create_dir(path)
	httpd:mkrdir(path)
end

function write_file_bylog(filename,body,mode)
	local fp = io.open(filename,mode)
	if fp == nil then
        return nil
    end
	fp:write(body)
	fp:flush()
	fp:close()
	return true
end

local function error_log()
	if httpd.status ~= 500 and httpd.status ~= 502 and httpd.status ~= 401 and httpd.status ~= 503 then return false end
	local logline = {localtime , ip , server_name , method , httpd.status , httpd.unparsed_uri , body_length , httpd.headers_in['referer'] , httpd.headers_in['user-agent'] , httpd.protocol,tms}
	local path = cpath .. '/logs/' .. server_name .. '/error'
	if not check_dir(path) then create_dir(path) end
	write_file_bylog(path .. '/'.. tostring(httpd.status) .. '.log',json.encode(logline) .. "\n",'ab')
end

local function exclude_extension()
	if not httpd.uri then return false end
	if not config['exclude_extension'] then return false end
	for _,v in ipairs(config['exclude_extension'])
	do
		if httpd:regex(httpd.uri,"[.]"..v.."$",0x01) then
			return true
		end
	end
	return false
end

local function exclude_url()
	if not httpd.uri then return false end
	if not config['exclude_url'] then return false end
	local the_uri = string.sub(httpd.unparsed_uri, 2)
	for _,v in pairs(config['exclude_url'])
	do
		local url = v["url"]
		local mode = v["mode"]
		if mode == "regular" then
			if httpd:regex(the_uri,url,0x01) then
				return true
			end
        else
			if the_uri == url then
				return true
			end
		end
	end
	return false
end

local function exclude_status()
	if not config["exclude_status"] then return false end
	local the_status = tostring(httpd.status)
	for _,v in ipairs(config['exclude_status'])
	do
		if the_status == v then
			return true
		end
	end
	return false
end

local function is_migrating()
	local file = io.open("/www/server/total/migrating", "rb")
	if file then return true end
	return false
end

function read_file_body_bylog(filename)
	local fp = io.open(filename,'rb')
	if not fp then
        return nil
    end
	fbody = fp:read("*a")
    fp:close()
    if fbody == '' then
        return nil
    end
	return fbody
end

local function split_bylog( str,reps )
    local resultStrList = {}
    string.gsub(str,'[^'..reps..']+',function(w)
        table.insert(resultStrList,w)
    end)
    return resultStrList
end

local function arrlen_bylog(arr)
	if not arr then return 0 end
	count = 0
	for _,v in ipairs(arr)
	do
		count = count + 1
	end
	return count
end

local function is_ipaddr_bylog(client_ip)
	local cipn = split_bylog(client_ip,'.')
	if arrlen_bylog(cipn) < 4 then return false end
	for _,v in ipairs({1,2,3,4})
	do
		local ipv = tonumber(cipn[v])
		if ipv == nil then return false end
		if ipv > 255 or ipv < 0 then return false end
	end
	return true
end

local function get_client_ip_bylog()
	local client_ip = "unknown"
	local cdn = config["cdn"]
	if cdn == true then
		for _,v in ipairs(config['cdn_headers'])
		do
			if request_header[v] ~= nil and request_header[v] ~= "" then
			    client_ip = split_bylog(httpd.headers_in[v],',')[1]
				break;
			end
		end
	end
	if type(client_ip) == 'table' then client_ip = "" end
	if string.match(client_ip,"^%d+%.%d+%.%d+%.%d+$") == nil or not is_ipaddr_bylog(client_ip) then
		client_ip = httpd.useragent_ip
		if client_ip == nil then
			client_ip = "unknown"
		end
	end
	return client_ip
end

local function arrip_bylog(ipstr)
	if ipstr == 'unknown' then return {0,0,0,0} end
	iparr = split_bylog(ipstr,'.')
	iparr[1] = tonumber(iparr[1])
	iparr[2] = tonumber(iparr[2])
	iparr[3] = tonumber(iparr[3])
	iparr[4] = tonumber(iparr[4])
	return iparr
end

function is_min_bylog(ip1,ip2)
	
	n = 0
	for _,v in ipairs({1,2,3,4})
	do
		if ip1[v] == ip2[v] then
			n = n + 1
		elseif ip1[v] > ip2[v] then
			break
		else
			return false
		end
	end
	return true
end

function is_max_bylog(ip1,ip2)
	n = 0
	for _,v in ipairs({1,2,3,4})
	do
		if ip1[v] == ip2[v] then
			n = n + 1
		elseif ip1[v] < ip2[v] then
			break
		else
			return false
		end
	end
	return true
end

function compare_ip_bylog(ips)
	if ip == 'unknown' then return true end
	if not is_max_bylog(ipn,ips[2]) then return false end
	if not is_min_bylog(ipn,ips[1]) then return false end
	return true
end

function get_length()
	local clen  = httpd.headers_out['Content-Length']
	if clen == nil then clen = 0 end
	if clen == 0 and httpd.filename then
		tmp = httpd:stat(httpd.filename)
      	if tmp then clen = tmp.size end
    end
	return tonumber(clen)
end


-- -- function write_logs_bylog()
-- --     local r_port = ':'..tostring(httpd.subprocess_env['REMOTE_PORT'])
-- --     if not httpd.subprocess_env['REMOTE_PORT'] then
-- --         r_port = ''
-- --     end
-- -- 	local logline = {localtime , ip .. r_port , server_name , httpd.method , httpd.status , httpd.unparsed_uri , body_length , httpd.headers_in['referer'] , httpd.headers_in['user-agent'],httpd.protocol,tms}
-- -- 	local path = cpath .. '/logs/' .. server_name
-- -- 	if not check_dir(path) then create_dir(path) end
-- -- 	write_file_bylog(path .. '/' .. today .. '.log',json.encode(logline) .. "\n",'ab')
-- -- end

function get_end_time()
	local s_time = os.time()
    local n_date = os.date("*t",s_time + 86400)
	n_date.hour = 0
	n_date.min = 0
	n_date.sec = 0
	d_time = os.time(n_date)
    return d_time - s_time
end

function get_data_on(filename)
	local data = cache:get(filename)
	if not data then
		data = read_file_body_bylog(filename)
		if not data then 
			data = {}
		else
			data = json.decode(data)
		end
	else
		data = json.decode(data)
	end
	return data
end

-- function save_data_on(filename,data)
-- 	local newdata = json.encode(data)
-- 	local extime = 0;
-- 	if string.find(filename,"%d+-%d+-%d+") then extime = 88400 end
-- 	cache:set(filename,newdata,extime)
-- 	if not cache:get(filename..'_lock') then
-- 		cache:set(filename..'_lock',1,10)
-- 		write_file_bylog(filename,newdata,'wb')
-- 	end
-- end


-- local function get_request_time()
-- 	return -1
-- end

local function get_spiders()
    return require "spiders"
end

local function match_spider()
	-- TODO 加入IP区间判断是否是蜘蛛请求
	-- 匹配蜘蛛请求
	local ua = ''
	if httpd.headers_in['user-agent'] then
		ua = httpd.headers_in['user-agent']
	end
	if not ua then
		return false, nil, 0
	end
	local is_spider = false
	local spider_name = nil
	
	local spider_table = {
		["baidu"] = 1,
		["bing"] = 2,
		["qh360"] = 3,
		["google"] = 4,
		["bytes"] = 5,
		["sogou"] = 6,
		["youdao"] = 7,
		["soso"] = 8,
		["dnspod"] = 9,
		["yandex"] = 10,
		["yisou"] = 11,
		["other"] = 12
	}

	local res,err = httpd:regex(ua, "(Baiduspider|Bytespider|360Spider|Sogou web spider|Sosospider|Googlebot|bingbot|AdsBot-Google|Google-Adwords|YoudaoBot|Yandex|DNSPod-Monitor|YisouSpider)", 0x01)
	if res then
		is_spider = true
		spider_match = string.lower(res[0])
		if string.find(spider_match, "baidu") then
			spider_name = "baidu"
		elseif string.find(spider_match, "bytes") then
			spider_name = "bytes"
		elseif string.find(spider_match, "360") then
			spider_name = "qh360"
		elseif string.find(spider_match, "sogou") then
			spider_name = "sogou"
		elseif string.find(spider_match, "soso") then
			spider_name = "soso"
		elseif string.find(spider_match, "google") then
			spider_name = "google"
		elseif string.find(spider_match, "bingbot") then
			spider_name = "bing"
		elseif string.find(spider_match, "youdao") then
			spider_name = "youdao"
		elseif string.find(spider_match, "dnspod") then
			spider_name = "dnspod"
		elseif string.find(spider_match, "yandex") then
			spider_name = "yandex"
		elseif string.find(spider_match, "yisou") then
			spider_name = "yisou"
			-- elseif string.find(spider_match, "yahoo") then
			-- 	spider_name = "yahoo"
		end	
	end

	if is_spider then 
		return is_spider, spider_name, spider_table[spider_name]
	end
	local other_res, err = httpd:regex(ua, "(Yahoo|DuckDuckGo)", 0x01)
	if other_res then
		-- debug("Other Spider:"..other_res[0].."/User Agent:"..tostring(ua))
		return true, "other", spider_table["other"]
	end

	return false, nil, 0
end

local function statistics_ipc()
	local ipc = 0 
	local ip_token = server_name..'_'..ip
	if not cache:get(ip_token) then
		ipc = 1
		cache:set(ip_token,1,get_end_time())
	end
	return ipc
end

local function statistics_request()
	-- 计算pv uv
	local pvc = 0
	local uvc = 0
	if method == 'GET' and tonumber(httpd.status) == 200 and body_length > 200 then
		local ua = ''
		if httpd.headers_in['user-agent'] then
			ua = string.lower(httpd.headers_in['user-agent'])
		end

		if httpd.headers_out['content-type'] then
			if string.find(httpd.headers_out['content-type'],'text/html') then
			-- if not httpd:regex(httpd.uri,"[.](js|css|png|jpeg|jpg|gif)$",0x01) then
				-- 判断IP是否重复的时间限定范围是请求的当前时间+24小时
				pvc = 1
				
				if ua then
					if string.find(ua,'mozilla') then
						local _today = os.date("%Y-%m-%d")
						local uv_token = httpd:md5(ip .. httpd.headers_in['user-agent'] .. _today)
						if not cache:get(uv_token) then
							uvc = 1
							cache:set(uv_token,1,get_end_time())
						end
					end
				end
			end
		end
	end
	return pvc, uvc
end

local function get_clients()
	-- local clients_json = cpath .. '/config/clients.json'
	-- local clients = get_data_on(clients_json)
	return require "clients"
end

local function match_pc_client(ua, clients)
	local msie = nil
	local msie_pattern = nil
	local safari = nil
	local safari_pattern = nil
	local chrome = nil
	local chrome_pattern = nil
	for column, pattern in pairs(clients) do
		if column == "safari" then
			safari_pattern = pattern
		elseif column == "chrome" then
			chrome_pattern = pattern
		elseif column == "msie" then
			msie_pattern = pattern
		elseif httpd:regex(ua, pattern, 0x01) then
			return column
		end
	end
	if msie_pattern and httpd:regex(ua, msie_pattern, 0x01) then
		return "msie"
	end
	if chrome_pattern and httpd:regex(ua, chrome_pattern, 0x01) then
		return "chrome"
	end
	if safari_pattern and httpd:regex(ua, safari_pattern, 0x01) then
		return "safari"
	end
	return nil
end

local function match_norepeat_client(ua, clients)
	if clients == nil then return end
	for name, pattern in pairs(clients) do
		if httpd:regex(ua, pattern, 0x01) then
			return name
		end
	end
end

local function get_update_field(field, value)
	return field.."="..field.."+"..tostring(value)
end

local function match_client()
	-- 匹配客户端
	local ua = ''
	if request_header['user-agent'] then
		ua = request_header['user-agent']
	end
	if not ua then
		return false, nil
	end
	local client_stat_fields = ""
	local clients_map = {
		["android"] = "android",
		["iphone"] = "iphone",
		["ipod"] = "iphone",
		["ipad"] = "iphone",
		["firefox"] = "firefox",
		["msie"] = "msie",
		["trident"] = "msie",
		["360se"] = "qh360",
		["360ee"] = "qh360",
		["360browser"] = "qh360",
		["qihoo"] = "qh360",
		["the world"] = "theworld",
		["theworld"] = "theworld",
		["tencenttraveler"] = "tt",
		["maxthon"] = "maxthon",
		["opera"] = "opera",
		["qqbrowser"] = "qq",
		["ucweb"] = "uc",
		["ubrowser"] = "uc",
		["safari"] = "safari",
		["chrome"] = "chrome",
		["metasr"] = "metasr",
		["2345explorer"] = "pc2345",
		["edge"] = "edeg",
		["edg"] = "edeg",
		["windows"] = "windows",
		["linux"] = "linux",
		["macintosh"] = "mac",
		["mobile"] = "mobile"
	}
	local mobile_regx = "(Mobile|Android|iPhone|iPod|iPad)"
	local mobile_res = httpd:regex(ua, mobile_regx, 0x01)
	--mobile
	if mobile_res then
		client_stat_fields = client_stat_fields..","..get_update_field("mobile", 1)
		mobile_res = string.lower(mobile_res[0])
		if mobile_res ~= "mobile" then
			client_stat_fields = client_stat_fields..","..get_update_field(clients_map[mobile_res], 1)
		end
	else
		--pc
		-- 匹配结果的顺序，与ua中关键词的顺序有关
		-- lua的正则不支持|语法
		-- 短字符串string.find效率要比ngx正则高
		local pc_regx1 = "(360SE|360EE|360browser|Qihoo|TheWorld|TencentTraveler|Maxthon|Opera|QQBrowser|UCWEB|UBrowser|MetaSr|2345Explorer|Edg[e]*)"
		local pc_res = httpd:regex(ua, pc_regx1, 0x01)
		local cls_pc = nil
		if not pc_res then
			if string.find(ua, "[Ff]irefox") then
				cls_pc = "firefox"
			elseif string.find(ua, "MSIE") or string.find(ua, "Trident") then
				cls_pc = "msie"
			elseif string.find(ua, "[Cc]hrome") then
				cls_pc = "chrome"
			elseif string.find(ua, "[Ss]afari") then
				cls_pc = "safari"
			end
		else
			cls_pc = string.lower(pc_res[0])
		end
		-- debug("UA:"..ua)
		-- debug("Cls pc:"..cls_pc)
		if cls_pc then
			client_stat_fields = client_stat_fields..","..get_update_field(clients_map[cls_pc], 1)
		else
			-- machine and other
			local machine_res, err = httpd:regex(ua, "([Cc]url|HeadlessChrome|[a-zA-Z]+[Bb]ot|[Ww]get|[Ss]pider|[Cc]rawler|[Ss]crapy|zgrab|[Pp]ython|java)", 0x01)
			if machine_res then
				client_stat_fields = client_stat_fields..","..get_update_field("machine", 1)
			else
				-- 移动端+PC端+机器以外 归类到 其他
				client_stat_fields = client_stat_fields..","..get_update_field("other", 1)
			end
		end

		local os_regx = "(Windows|Linux|Macintosh)"
		local os_res = httpd:regex(ua, os_regx, 0x01)
		if os_res then
			os_res = string.lower(os_res[0])
			client_stat_fields = client_stat_fields..","..get_update_field(clients_map[os_res], 1)
		end
	end

	local other_regx = "MicroMessenger"
	local other_res = string.find(ua, other_regx)
	if other_res then
		client_stat_fields = client_stat_fields..","..get_update_field("weixin", 1)
	end
	if client_stat_fields then
		client_stat_fields = string.sub(client_stat_fields, 2)
	end
	return client_stat_fields
end

local function total_flow()
	local flow_key = server_name .. '_flow'
	local flow_write = server_name .. '_flow_write'
	if cache:get(flow_write) then
		cache:incr(flow_key,body_length)
	else
		local write_time = now_timestamp
		local now_flow = cache:get(flow_key)
		if not now_flow then now_flow = 0 end
		now_flow = now_flow + body_length
		local realtime_data = tostring(now_flow)..","..tostring(write_time)
		-- debug("Current realtime data:"..tostring(realtime_data))

		local queue_length = 10
		local position_key = flow_key.."_positon"
		local queue_key = flow_key.."_queue_"

		local req_position = 0
		if cache:get(position_key) == nil then
			cache:set(position_key, 1)
			req_position = 1
		else
			cache:incr(position_key, 1)
			req_position = cache:get(position_key)
			if req_position >= queue_length then
				cache:set(position_key, 1)
				req_position = 1
			end
		end

		cache:set(queue_key..tostring(req_position), realtime_data)

		local data = ""
		for i=1, queue_length, 1 do
			local _d = cache:get(queue_key..tostring(i))
			if _d ~= nil then
				data = data.."\n".._d 
			end
		end

		local path = cpath .. 'logs/' .. server_name
		if not check_dir(path) then create_dir(path) end

		local filename = path .. '/flow_sec.json'
		write_file_bylog(filename,tostring(data),'w+')
		-- debug("All realtime data:"..tostring(data))

		cache:set(flow_key,body_length,2)
		cache:set(flow_write,1,1)
	end
end

local function total_req_sec()
	local _key = server_name .. '_req'
	local _write = server_name .. '_req_write'
	if cache:get(_write) then
		cache:incr(_key,1)
	else
		local write_time = now_timestamp
		local now_ = cache:get(_key)
		if not now_ then now_ = 0 end
		now_ = now_ + 1
		local realtime_data = tostring(now_)..","..tostring(write_time)

		local queue_length = 10
		local position_key = _key.."_positon"
		local queue_key = _key.."_queue_"

		local req_position = 0
		if cache:get(position_key) == nil then
			cache:set(position_key, 1)
			req_position = 1
		else
			cache:incr(position_key, 1)
			req_position = cache:get(position_key)
			if req_position >= queue_length then
				cache:set(position_key, 1)
				req_position = 1
			end
		end

		cache:set(queue_key..tostring(req_position), realtime_data)

		local data = ""
		for i=1, queue_length, 1 do
			local _d = cache:get(queue_key..tostring(i))
			if _d ~= nil then
				data = data.."\n".._d 
			end
		end

		local path = cpath .. 'logs/' .. server_name
		if not check_dir(path) then create_dir(path) end
		local filename = path .. '/req_sec.json'
		local write_res = write_file_bylog(filename,data,'w+')

		cache:set(_key,1,2)
		cache:set(_write,1,1)
	end
end

local function get_domain()
	local domain = request_header['host']
	if domain ~= nil then
		domain = string.gsub(domain, "_", ".")
	else
		domain = "unknown"
	end
	return domain
end


local function is_storing(store_server_name)
	local store_status = cache:get(store_server_name.."_STORING")
	if store_status ~= nil and store_status == true then
		return true 
	end
	return false
end

local function get_store_key()
	return os.date("%Y%m%d%H", os.time())
end

local function update_stat(update_server_name, db, stat_table, key, columns)
	-- 根据指定表名，更新统计数据
	if not columns then return end
	local stmt = db:prepare(string.format("INSERT INTO %s(time) SELECT :time WHERE NOT EXISTS(SELECT time FROM %s WHERE time=:time);", stat_table, stat_table))
	stmt:bind_names{time=key}
	stmt:step()
	stmt:finalize()
	local update_sql = "UPDATE ".. stat_table .. " SET " .. columns
	update_sql = update_sql .. " WHERE time=" .. key
	status, errorString = db:exec(update_sql)
end

local function init_db(log_dir, server_name)
	-- 初始化日志db
	local path = log_dir .."/".. server_name
	if not check_dir(path) then create_dir(path) end
	local db_name = "logs.db"
	local db_path = path .. "/" .. db_name
	if cache:get(db_path) then
		return
	end
	local db = sqlite3.open(db_path)
	local table_name = "site_logs"
	local stmt = db:prepare("SELECT COUNT(*) FROM sqlite_master where type='table' and name=?")
	local rows = 0
	if stmt ~= nil then
		stmt:bind_values(table_name)
		stmt:step()
		rows = stmt:get_uvalues()
		stmt:finalize()
	end
	if stmt ~= nil and rows >0 then
		cache:set(db_path, true)
	else
		status, errorString = db:exec([[PRAGMA synchronous = 0]])
		status, errorString = db:exec([[PRAGMA page_size = 4096]])
		status, errorString = db:exec([[PRAGMA journal_mode = wal]])
		status, errorString = db:exec([[PRAGMA journal_size_limit = 1073741824]])
		status,errorString = db:exec[[
			CREATE TABLE site_logs (
				time INTEGER,
				ip TEXT,
				domain TEXT,
				server_name TEXT,
				method TEXT,
				status_code INTEGER,
				uri TEXT,
				body_length INTEGER,
				referer TEXT DEFAULT "",
				user_agent TEXT,
				is_spider INTEGER DEFAULT 0,
				protocol TEXT,
				request_time INTEGER,
				request_headers TEXT DEFAULT "",
				ip_list TEXT DEFAULT ""
			)]]
		status, errorString = db:exec([[CREATE INDEX time_inx ON site_logs(time)]])
		-- 客户端统计明细，按小时统计
		-- 客户端的定义: {client:"Android", column_name:"android"}
		status, errorString = db:exec[[
			CREATE TABLE client_stat(
				time INTEGER PRIMARY KEY,
				weixin INTEGER DEFAULT 0,
				android INTEGER DEFAULT 0,
				iphone INTEGER DEFAULT 0,
				mac INTEGER DEFAULT 0,
				windows INTEGER DEFAULT 0,
				linux INTEGER DEFAULT 0,
				edeg INTEGER DEFAULT 0,
				firefox INTEGER DEFAULT 0,
				msie INTEGER DEFAULT 0,
				metasr INTEGER DEFAULT 0,
				qh360 INTEGER DEFAULT 0,
				theworld INTEGER DEFAULT 0,
				tt INTEGER DEFAULT 0,
				maxthon INTEGER DEFAULT 0,
				opera INTEGER DEFAULT 0,
				qq INTEGER DEFAULT 0,
				uc INTEGER DEFAULT 0,
				pc2345 INTEGER DEFAULT 0, 
				safari INTEGER DEFAULT 0,
				chrome INTEGER DEFAULT 0,
				machine INTEGER DEFAULT 0,
				mobile INTEGER DEFAULT 0,
				other INTEGER DEFAULT 0
			)
		]]

        -- 请求分类统计明细，按小时统计
		-- time: 2021032500 时间刻度, 以小时为单位
		status, errorString = db:exec[[
				CREATE TABLE request_stat(
					time INTEGER PRIMARY KEY,
    				req INTEGER DEFAULT 0,
    				pv INTEGER DEFAULT 0,
    				uv INTEGER DEFAULT 0,
    				ip INTEGER DEFAULT 0,
    				length INTEGER DEFAULT 0,
                    spider INTEGER DEFAULT 0,
                    status_500 INTEGER DEFAULT 0, 
                    status_501 INTEGER DEFAULT 0, 
                    status_502 INTEGER DEFAULT 0, 
                    status_503 INTEGER DEFAULT 0, 
                    status_504 INTEGER DEFAULT 0, 
                    status_505 INTEGER DEFAULT 0, 
                    status_506 INTEGER DEFAULT 0, 
                    status_507 INTEGER DEFAULT 0, 
                    status_509 INTEGER DEFAULT 0, 
                    status_510 INTEGER DEFAULT 0, 
                    status_400 INTEGER DEFAULT 0, 
                    status_401 INTEGER DEFAULT 0, 
                    status_402 INTEGER DEFAULT 0, 
                    status_403 INTEGER DEFAULT 0, 
                    status_404 INTEGER DEFAULT 0, 
                    http_get INTEGER DEFAULT 0, 
                    http_post INTEGER DEFAULT 0, 
                    http_put INTEGER DEFAULT 0, 
                    http_patch INTEGER DEFAULT 0, 
                    http_delete INTEGER DEFAULT 0
				)
		]]

		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_405 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_406 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_407 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_408 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_409 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_410 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_411 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_412 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_413 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_414 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_415 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_416 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_417 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_418 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_421 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_422 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_423 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_424 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_425 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_426 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_449 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_451 INTEGER DEFAULT 0;")	
		-- 蜘蛛统计明细，按小时统计
		-- 按照已有蜘蛛分类，分别建立对应的统计列。
		status, errorString = db:exec[[
			CREATE TABLE spider_stat(
				time INTEGER PRIMARY KEY,
				bytes INTEGER DEFAULT 0,
				bing INTEGER DEFAULT 0,
				soso INTEGER DEFAULT 0,
				yahoo INTEGER DEFAULT 0,
				sogou INTEGER DEFAULT 0,
				google INTEGER DEFAULT 0,
				baidu INTEGER DEFAULT 0,
				qh360 INTEGER DEFAULT 0,
				youdao INTEGER DEFAULT 0,
				yandex INTEGER DEFAULT 0,
				dnspod INTEGER DEFAULT 0,
				yisou INTEGER DEFAULT 0,
				other INTEGER DEFAULT 0
			)
		]]
        
		local sql_str1 = [[
			CREATE TABLE uri_stat (
				uri_md5 CHAR(32) PRIMARY KEY,
				uri TEXT,
				day1 INTEGER DEFAULT 0,
				day2 INTEGER DEFAULT 0,
				day3 INTEGER DEFAULT 0,
				day4 INTEGER DEFAULT 0,
				day5 INTEGER DEFAULT 0,
				day6 INTEGER DEFAULT 0,
				day7 INTEGER DEFAULT 0,
				day8 INTEGER DEFAULT 0,
				day9 INTEGER DEFAULT 0,
				day10 INTEGER DEFAULT 0,
				day11 INTEGER DEFAULT 0,
				day12 INTEGER DEFAULT 0,
				day13 INTEGER DEFAULT 0,
				day14 INTEGER DEFAULT 0,
				day15 INTEGER DEFAULT 0,
				day16 INTEGER DEFAULT 0,
				day17 INTEGER DEFAULT 0,
				day18 INTEGER DEFAULT 0,
				day19 INTEGER DEFAULT 0,
				day20 INTEGER DEFAULT 0,
				day21 INTEGER DEFAULT 0,
				day22 INTEGER DEFAULT 0,
				day23 INTEGER DEFAULT 0,
				day24 INTEGER DEFAULT 0,
				day25 INTEGER DEFAULT 0,
				day26 INTEGER DEFAULT 0,
				day27 INTEGER DEFAULT 0,
				day28 INTEGER DEFAULT 0,
				day29 INTEGER DEFAULT 0,
				day30 INTEGER DEFAULT 0,
				day31 INTEGER DEFAULT 0,
				day32 INTEGER DEFAULT 0
			)]]
		status, errorString = db:exec(sql_str1)	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow1 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow2 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow3 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow4 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow5 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow6 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow7 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow8 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow9 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow10 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow11 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow12 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow13 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow14 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow15 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow16 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow17 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow18 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow19 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow20 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow21 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow22 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow23 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow24 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow25 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow26 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow27 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow28 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow29 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow30 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow31 INTEGER DEFAULT 0;")	

		local sql_str1 = [[
			CREATE TABLE ip_stat (
				ip CHAR(15) PRIMARY KEY,
				area CHAR(8) DEFAULT "",
				day1 INTEGER DEFAULT 0,
				day2 INTEGER DEFAULT 0,
				day3 INTEGER DEFAULT 0,
				day4 INTEGER DEFAULT 0,
				day5 INTEGER DEFAULT 0,
				day6 INTEGER DEFAULT 0,
				day7 INTEGER DEFAULT 0,
				day8 INTEGER DEFAULT 0,
				day9 INTEGER DEFAULT 0,
				day10 INTEGER DEFAULT 0,
				day11 INTEGER DEFAULT 0,
				day12 INTEGER DEFAULT 0,
				day13 INTEGER DEFAULT 0,
				day14 INTEGER DEFAULT 0,
				day15 INTEGER DEFAULT 0,
				day16 INTEGER DEFAULT 0,
				day17 INTEGER DEFAULT 0,
				day18 INTEGER DEFAULT 0,
				day19 INTEGER DEFAULT 0,
				day20 INTEGER DEFAULT 0,
				day21 INTEGER DEFAULT 0,
				day22 INTEGER DEFAULT 0,
				day23 INTEGER DEFAULT 0,
				day24 INTEGER DEFAULT 0,
				day25 INTEGER DEFAULT 0,
				day26 INTEGER DEFAULT 0,
				day27 INTEGER DEFAULT 0,
				day28 INTEGER DEFAULT 0,
				day29 INTEGER DEFAULT 0,
				day30 INTEGER DEFAULT 0,
				day31 INTEGER DEFAULT 0,
				day32 INTEGER DEFAULT 0
			)]]
		status,errorString = db:exec(sql_str1)
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow1 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow2 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow3 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow4 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow5 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow6 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow7 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow8 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow9 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow10 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow11 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow12 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow13 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow14 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow15 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow16 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow17 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow18 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow19 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow20 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow21 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow22 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow23 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow24 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow25 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow26 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow27 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow28 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow29 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow30 INTEGER DEFAULT 0;")	
		status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow31 INTEGER DEFAULT 0;")
		
		status, errorString = db:exec[[
			CREATE TABLE referer_stat(
				time INTEGER,
    			domain TEXT,
                count INTEGER DEFAULT 0
			)
		]]

		cache:set(db_path, true)
		debug("初始化日志数据库完成。")
	end
	db:close()
end

function get_server_name()
	local c_name = httpd.server_name
	local my_name = cache:get(c_name)
	if my_name then return my_name end
	-- local tmp = read_file_body_bylog(cpath .. '/domains.json')
	-- if not tmp then return c_name end
	-- local domains = json.decode(tmp)
	local domains = require "domains"
	for _,v in ipairs(domains)
	do
		if c_name == v["name"] then
	        cache:set(c_name, v['name'],3600)
			return v["name"]
		end
		for _,d_name in ipairs(v['domains'])
		do
			if string.find(d_name, "*") then
	            new_domain = string.gsub(d_name, "*", ".*")
	            if string.find(c_name, new_domain) then
	                cache:set(c_name, v['name'],3600)
	                return v['name']
	            end
            elseif c_name == d_name then
				cache:set(c_name,v['name'],3600)
				return v['name']
			end
		end
	end

	local tconf = httpd:activeconfig()
	if not tconf then 
		return no_binding_server_name 
	end
	local tmp = split_bylog(tconf[1].file,'/')
	if not tmp then 
		return no_binding_server_name
	end
	return string.gsub(tmp[arrlen_bylog(tmp)],'.conf$','')
end

local function get_config(site, key)

	local config_data = total_config
	-- local config_data = json.decode(read_file_body_bylog(cpath..'/config.json'))
	local config = nil
	if config_data["global"] == nil then return nil end
	global_config = config_data["global"]
	if key == "monitor" then
		if not global_config["monitor"] then
			return global_config["monitor"]
		end
	end
	if config_data[server_name] == nil then
		config = global_config
	else
		config = config_data[server_name]
		for k, v in pairs(global_config) do
			if config[k] == nil then
				config[k] = v
			end
		end
	end
    return config
end

local function cache_value(id, key, value)
	local line_id_key = "logline_"..server_name.."_"..id.."_"..key
	cache:set(line_id_key, value)
end

local function clear_value(store_server_name, id, key)
	local line_id_key = "logline_"..store_server_name.."_"..id.."_"..key
	cache:delete(line_id_key)
end

local function get_value(store_server_name, id, key)
	local line_id_key = "logline_"..store_server_name.."_"..id.."_"..key
	local value = cache:get(line_id_key)
	-- cache:delete(line_id_key)
	return value
end

local function load_update_day(update_server_name)
	local _file = cpath.."logs/"..update_server_name.."/update_day.log"
	return read_file_body_bylog(_file)
end

local function write_update_day(update_server_name)
	local update_day = today
	local _file = cpath.."logs/"..update_server_name.."/update_day.log"
	write_file_bylog(_file, update_day, "w")
end

local function statistics_uri(db, stat_server_name, uri, uri_md5, body_length)
	-- 统计uri请求次数和流量
	local open_statistics_uri = config["statistics_uri"]
	if not open_statistics_uri then return true end

	local stat_sql = nil

	stat_sql = "INSERT INTO uri_stat(uri_md5,uri) SELECT \""..uri_md5.."\",\""..uri.."\" WHERE NOT EXISTS (SELECT uri_md5 FROM uri_stat WHERE uri_md5=\""..uri_md5.."\");"
	local res, err = db:exec(stat_sql)
	-- debug("Insert uri stat res:"..tostring(res))
	-- debug("Insert uri stat err:"..tostring(err))
	
	if is_spider then
		stat_sql = "UPDATE uri_stat SET "..day_column.."="..day_column.."+1,"..spider_column.."="..spider_column.."+"..body_length.." WHERE uri_md5=\""..uri_md5.."\""
	else
		stat_sql = "UPDATE uri_stat SET "..day_column.."="..day_column.."+1,"..flow_column.."="..flow_column.."+"..body_length.." WHERE uri_md5=\""..uri_md5.."\""
	end
	local res, err = db:exec(stat_sql)
	-- debug("Stat sql:"..stat_sql)
	-- debug("Update uri stat res:"..tostring(res))
	-- debug("Update uri stat err:"..tostring(err))
	
	return true
	-- local stmt = db:prepare(string.format("INSERT INTO %s(time) SELECT :time WHERE NOT EXISTS(SELECT time FROM %s WHERE time=:time);", stat_table, stat_table))
end

local function statistics_ip(db, stat_server_name, ip, body_length)
	local open_statistics_ip = config["statistics_ip"]	
	if not open_statistics_ip then return true end
	
	local stat_sql = nil

	stat_sql = "INSERT INTO ip_stat(ip) SELECT \""..ip.."\" WHERE NOT EXISTS (SELECT ip FROM ip_stat WHERE ip=\""..ip.."\");"
	local res, err = db:exec(stat_sql)
	-- debug("Ip insert stat res:"..tostring(res))
	-- debug("Ip insert stat err:"..tostring(err))
	
	if is_spider then
		stat_sql = "UPDATE ip_stat SET "..day_column.."="..day_column.."+1,"..spider_column.."="..spider_column.."+"..body_length.." WHERE ip=\""..ip.."\""
	else
		stat_sql = "UPDATE ip_stat SET "..day_column.."="..day_column.."+1,"..flow_column.."="..flow_column.."+"..body_length.." WHERE ip=\""..ip.."\""
	end

	local res, err = db:exec(stat_sql)
	-- debug("Update ip res:"..tostring(res))
	-- debug("Update ip err:"..tostring(err))
	return true
end


-- 存储单行日志
local function store_line(db, stmt, store_server_name, lineno)

	local logvalue = get_value(store_server_name, lineno, "logline")
	if not logvalue then return false end

	local logline = json.decode(logvalue)
	local time = logline["time"]
	local time_key = logline["time_key"]
	local ip = logline["ip"]
	local ip_list = logline["ip_list"]
	local domain = logline["domain"]
	local server_name = logline["server_name"]
	local real_server_name = logline["real_server_name"]
	local method = logline["method"]
	local status_code = logline["status_code"]
	local uri = logline["uri"]
	local request_uri = logline["request_uri"]
	if not request_uri then
		request_uri = uri
	end
	local body_length = logline["body_length"]
	local referer = logline["referer"]
	local user_agent = logline["user_agent"]
	local protocol = logline["protocol"]
	local request_time = logline["request_time"]
	local is_spider = logline["is_spider"]
	local request_headers = logline["request_headers"]
	local excluded = logline["excluded"]
	local id = logline["id"]

	local request_stat_fields = nil 
	local client_stat_fields = nil
	local spider_stat_fields = nil
	local stat_fields = get_value(store_server_name, id, "STAT_FIELDS")
	if stat_fields == nil then
		debug("Log stat fields is nil.")
		debug("Logdata:"..logvalue)
	else
		stat_fields = split_bylog(stat_fields, ";")
		request_stat_fields = stat_fields[1]
		client_stat_fields = stat_fields[2]
		spider_stat_fields = stat_fields[3]

		if "x" == client_stat_fields then
			client_stat_fields = nil
		end

		if "x" == spider_stat_fields then
			spider_stat_fields = nil
		end
	end

	-- debug("Request stat fields:"..tostring(request_stat_fields))
	-- debug("Spider stat fields:"..tostring(spider_stat_fields))
	-- debug("Client stat fields:"..tostring(client_stat_fields))

	if not excluded then
		stmt:bind_names{
			time=time,
			ip=ip,
			domain=domain,
			server_name=real_server_name,
			method=method,
			status_code=status_code,
			uri=request_uri,
			body_length=body_length,
			referer=referer,
			user_agent=user_agent,
			protocol=protocol,
			request_time=request_time,
			is_spider=is_spider,
			request_headers=request_headers,
			ip_list=ip_list
		}
		local res, err = stmt:step()
		if tostring(res) == "5" then
			return false
		end
		stmt:reset()
		update_stat(store_server_name, db, "client_stat", time_key, client_stat_fields)
		update_stat(store_server_name, db, "spider_stat", time_key, spider_stat_fields)
		if is_spider == 0 then
			local ok, err = pcall(function() statistics_uri(db, store_server_name, uri, httpd:md5(uri), body_length) end)
			-- debug("Statistics api result:"..tostring(ok))
			-- debug("Statistics api err:"..tostring(err))
			local ok, err = pcall(function() statistics_ip(db, store_server_name, ip, body_length) end)
		end
		-- debug("Statistics ip result:"..tostring(ok))
		-- debug("Statistics ip err:"..tostring(err))
	end
	update_stat(store_server_name, db, "request_stat", time_key, request_stat_fields)
	return true
end

local function store_logs(store_server_name)
	if is_migrating() == true then 
		-- debug("Migrating...")
		return
	end
	local flush_data = false
	local waiting_store_key = store_server_name .. "_WAITING_STORE"
	local flush_data_key = store_server_name .. "_FLUSH_DATA"
	local cache_count_id_key = store_server_name.."_CACHE_COUNT"
	local cache_count = cache:get(cache_count_id_key)
	if not cache_count then cache_count = 0 end
	-- debug("if:"..tostring(cache:get(waiting_store_key) and cache_count < cache_count_threshold))
	if not cache:get(waiting_store_key) or cache_count>cache_count_threshold then 
		local x = "go"
    else
		if cache:get(flush_data_key) then
			flush_data = true
		else
			return
		end
	end

	if is_storing(store_server_name) == true then 
		-- debug("其他worker正在存储中，稍候存储。")
		cache:delete(flush_data_key)
		return 
	end

	-- debug("Cache count:"..cache_count)
	-- debug("flush data:"..tostring(flush_data))
	local storing_key = store_server_name.."_STORING"
	cache:set(storing_key,true, 60)
	if flush_data == false then
		cache:set(waiting_store_key, os.time(), cache_interval_threshold)
	end
	cache:delete(flush_data_key)

	-- 开始存储数据
	local last_insert_id_key = store_server_name.."_LAST_INSERT_ID"
	local store_start_id_key = store_server_name.."_STORE_START"
	-- 1.计算存储数据区间段
	local last_id = tonumber(cache:get(last_insert_id_key))
	local store_start = tonumber(cache:get(store_start_id_key))
	-- debug("store start:"..tostring(store_start))
	-- debug("flush data:"..tostring(flush_data))
	if store_start == nil then
		store_start = 1
	end
	local store_end = last_id
	if not store_end then
		store_end = 1
	end

	local site_config = get_config(store_server_name)


	local store_count = 0
	-- local store_start_time = socket.gettime() * 1000
	-- debug("start:"..tostring(store_start).."/end:"..tostring(store_end))
	local stmt2 = nil
	local log_dir = cpath .. 'logs/'
	data_dir = site_config["data_dir"] 
	if data_dir then
		log_dir = data_dir
	end
	local db_path = log_dir .. "/" .. store_server_name .. "/logs.db"
	if not cache:get(db_path) then
		init_db(log_dir, store_server_name)
	end
	if not file_exists(db_path) then
		cache:set(db_path, nil)
		init_db(log_dir, store_server_name)
	end
	local db = sqlite3.open(db_path)
	if db ~= nil then
		stmt2 = db:prepare[[INSERT INTO site_logs(
			time, ip, domain, server_name, method, status_code, uri, body_length,
			referer, user_agent, protocol, request_time, is_spider, request_headers, ip_list)
			VALUES(:time, :ip, :domain, :server_name, :method, :status_code, :uri,
			:body_length, :referer, :user_agent, :protocol, :request_time, :is_spider,
			:request_headers, :ip_list)]]
	end

	if db == nil or stmt2 == nil then
		debug("网站监控报表数据库连接繁忙或者出现异常。")
		cache:set(storing_key, false)

		if db and db:isopen() then
			db:close()
		end
		return true
	end
	-- local start_transaction_time = socket.gettime()*1000 
	status, errorString = db:exec([[BEGIN TRANSACTION]])

	update_day = load_update_day(store_server_name)
	if not update_day or update_day ~= today then
		-- reset
		local update_sql = "UPDATE uri_stat SET "..day_column.."=0,"..spider_column.."=0,"..flow_column.."=0"
		status, errorString = db:exec(update_sql)

		update_sql = "UPDATE ip_stat SET "..day_column.."=0,"..flow_column.."=0"
		status, errorString = db:exec(update_sql)
		write_update_day(store_server_name)
	end

	if store_end >= store_start then
		for i=store_start, store_end, 1 do
			if store_line(db, stmt2, store_server_name, i) then
				store_count = store_count + 1
				clear_value(store_server_name, i, "logline")
				clear_value(store_server_name, i, "STAT_FIELDS")
				-- if store_count >= cache_count_threshold then
				-- 	store_end = i
				-- 	break
				-- end
			end
		end
	else
		local _tmp_store_end = store_end
		store_end = max_log_id
		for i=store_start, store_end, 1 do
			if store_line(db, stmt2, store_server_name, i) then
				store_count = store_count + 1
				clear_value(store_server_name, i, "logline")
				clear_value(store_server_name, i, "STAT_FIELDS")
			end
		end

		store_start = 1
		store_end = _tmp_store_end
		for i=store_start, store_end, 1 do
			if store_line(db, stmt2, store_server_name, i) then
				store_count = store_count + 1
				clear_value(store_server_name, i, "logline")
				clear_value(store_server_name, i, "STAT_FIELDS")
			end
		end
	end

	local res, err = stmt2:finalize()
	if tostring(res) == "5" then
		debug("Finalize 数据库繁忙，稍候存储。")
	else
		local now_date = os.date("*t")
		local _save_day = site_config["save_day"]
		local save_date_timestamp = os.time{year=now_date.year,
			month=now_date.month, day=now_date.day-_save_day, hour=0}
		local exec_res = db:exec("DELETE FROM site_logs WHERE time<"..tostring(save_date_timestamp))

		db:execute([[COMMIT]])

		cache:set(store_start_id_key, store_end)
		cache:decr(cache_count_id_key, store_count)
		local cache_count = cache:get(cache_count_id_key)
		if not cache_count then cache_count = 0 end
		if cache_count > cache_count_threshold or cache_count < 0 then
			cache:delete(cache_count_id_key)
		end
	end

	if db and db:isopen() then
		db:close()
	end

	cache:set(waiting_store_key, os.time(), cache_interval_threshold)
	cache:set(storing_key, false)
end

local function cache_logs()

	local excluded = false
	excluded = exclude_status() or exclude_extension() or exclude_url()

	-- if excluded then return apache2.OK end
	-- excluded = false
	-- debug("Excluded:"..tostring(excluded))

	domain = get_domain()
	method = httpd.method
	if method == "" or not method then return apache2.OK end
	if not httpd.notes['RT'] then
	    tms = 1
	else
	    tms = split_bylog(tostring((httpd:clock() - httpd.notes['RT']) / 1000),'.')[1]
	end
	
	ip = get_client_ip_bylog()
	local ip_list = httpd.headers_in['x-forwarded-for'] 
	if ip and not ip_list then
		ip_list = ip
	end
	-- ipn = arrip_bylog(ip)

	local last_insert_id_key = server_name .. "_LAST_INSERT_ID"
	local generating_id_key = server_name .. "GENERATING"
	local cache_count_id_key = server_name .. "_CACHE_COUNT"
	if not cache:get(server_name.."_CLEAR_CACHE_COUNT") then
		cache:delete(server_name.."_CACHE_COUNT")
		cache:set(server_name.."_CLEAR_CACHE_COUNT", true)
		-- debug("Cleared cache count.")
	end
	if not cache:get(cache_count_id_key) then
		cache:set(cache_count_id_key, 1)
	else
		cache:incr(cache_count_id_key, 1)
	end
	
	local new_id = nil
	if not cache:get(last_insert_id_key) then
		cache:set(last_insert_id_key, 1)
		new_id = 1
	else
		new_id = tonumber(cache:incr(last_insert_id_key, 1))
		if new_id >= max_log_id then
			cache:set(last_insert_id_key, 1)
			new_id = 1
		end
	end
	-- debug("Store count2:"..tostring(cache:get(cache_count_id_key)))

	-- debug("New id:"..tostring(new_id))
	local request_time = tms
	local status_code = httpd.status
	local referer = httpd.headers_in['referer'] 
	local request_uri = httpd.unparsed_uri
	local uri = httpd.uri
	local protocol = httpd.protocol
	local is_spider = 0
	local time_key = get_store_key()
	local user_agent = httpd.headers_in['user-agent']
	local real_server_name = server_name
	if server_name == no_binding_server_name then
		real_server_name = httpd.server_name
	end
		
    local logline = {
	    id=new_id,
		time_key=time_key,
		time=os.time(), 
		ip=ip, 
		domain=domain,
		server_name=server_name, 
		real_server_name=real_server_name,
		method=method, 
		status_code=status_code, 
		uri=uri, 
		request_uri=request_uri,
		body_length=body_length, 
		referer=referer, 
		user_agent=request_header['user-agent'], 
		protocol=protocol,
		is_spider=0,
		request_time=request_time,
		excluded=excluded,
		request_headers="",
		ip_list=ip_list
    }

	local request_stat_fields = "req=req+1,length=length+"..body_length
	local spider_stat_fields = "x"
	local client_stat_fields = "x"
	local is_spider = false
	local request_spider = nil
	local spider_index = 0
	-- debug("Excluded:"..tostring(excluded).."/uri:"..httpd.uri)
	if not excluded then
		-- if status_code == 500 or (method=="POST" and config["record_post_args"]==true) then
		-- 	local data = ""
		-- 	local ok, err = pcall(function() data=get_http_original() end)
		-- 	if ok and not err then
		-- 		logline["request_headers"] = data
		-- 	end
		-- 	debug("Get http orgininal ok:"..tostring(ok))
		-- 	debug("Get http orgininal res:"..tostring(data))
		-- 	-- cache_value(new_id, "request_headers", ngx.req.raw_header())
		-- end

		if string.find("500,501,502,503,504,505,506,507,509,510,400,401,402,403,404,405,406,407,408,409,410,411,412,413,414,415,416,417,418,421,422,423,424,425,426,449,451", status_code) then
			local field = "status_"..status_code
			request_stat_fields = request_stat_fields .. ","..field.."="..field.."+1"
		end
		local lower_method = string.lower(method)
		if string.find("get,post,put,patch,delete", lower_method) then
			local field = "http_"..lower_method
			request_stat_fields = request_stat_fields .. ","..field.."="..field.."+1"
		end

     	local ipc = 0
     	local pvc = 0
     	local uvc = 0
		is_spider, request_spider, spider_index = match_spider()
		if not is_spider then
			client_stat_fields  = match_client()	
			-- debug("Client stat fields:"..tostring(client_stat_fields))
			if #client_stat_fields == 0 then
				client_stat_fields = request_stat_fields..",other=other+1"
			else
				if string.find(client_stat_fields, "machine") then
					if statistics_machine_access then
						pvc, uvc = statistics_request()
					end	
				else
					pvc, uvc = statistics_request()
				end
			end
			ipc = statistics_ipc()
		else
			logline["is_spider"] = spider_index
			local field = "spider"
			request_stat_fields = request_stat_fields .. ","..field.."="..field.."+"..1
			spider_stat_fields = request_spider.."="..request_spider.."+"..1
		end
		if ipc > 0 then 
			request_stat_fields = request_stat_fields..",ip=ip+1"
		end
		if uvc > 0 then 
			request_stat_fields = request_stat_fields..",uv=uv+1"
		end
		if pvc > 0 then
			request_stat_fields = request_stat_fields..",pv=pv+1"
		end
		-- debug("Is spider:"..tostring(is_spider))
	end
	cache_value(new_id, "STAT_FIELDS", request_stat_fields..";"..client_stat_fields..";"..spider_stat_fields)
	cache_value(new_id, "logline", json.encode(logline))
	-- debug("Cached.")
end

function set_request_time(r)
    r.notes['RT'] = r:clock()
    return apache2.DECLINED
end

function run_logs(request)
	httpd = request
	if httpd.uri == '/favicon.ico' or httpd.status == 0 or httpd.status == 416 then return apache2.OK end
	local presult, err = pcall(
		function() 
			json = require "cjson" 
			memcached = require "memcached"
			socket = require "socket"
			sqlite3 = require "lsqlite3"
			total_config = require "total_config"
		end
	) 
	
	if not presult then 
		debug("引入依赖出现错误:"..tostring(err))
		return apache2.OK
	end

    cache = memcached.Connect("127.0.0.1", 11211)
	if not cache then
		cache = memcached.Connect("localhost", 11211)
	end
	if not cache then return apache2.OK end

	no_binding_server_name = "btunknownbt"
	server_name = string.gsub(get_server_name(),'_','.')
	if server_name == "127.0.0.1" or server_name == 'phpinfo' then return apache2.OK end
	-- debug("Server name:"..server_name)
	-- debug("uri:"..httpd.uri)

    config = get_config(server_name)
	if config == nil then return true end
	monitor = config["monitor"]
	if monitor == false then
		-- debug("------------------------站点:"..server_name.."关闭监控----------------------")
		return apache2.OK
	end

	max_log_id = 99999999999999

	today = os.date("%Y%m%d")
	day = os.date("%d")
	number_day = tonumber(day)
	day_column = "day"..number_day
	flow_column = "flow"..number_day
	spider_column = "spider_flow"..number_day

	request_header = httpd.headers_in
	out_header = httpd.headers_out

	if server_name == "httpd-vhosts" then
		return apache2.OK
	end

	-- debug("Monitor status:"..tostring(monitor))
	-- error_log()
	
	body_length = get_length()
	now_timestamp = os.time(os.date("*t"), os.time())

	total_flow()
	total_req_sec()

	statistics_machine_access = config["statistics_machine_access"]
	-- debug("Statistics machine access:"..tostring(statistics_machine_access))

	cache_logs()
	store_logs(server_name)
	return apache2.OK
end
