log_by_lua_block {
	local version = "1.5"
	-- local ok, app = pcall(require, "lsqlite3")
	local cpath = "/www/server/total/"
    if not package.cpath:find(cpath) then
        package.cpath = cpath .. "?.so;" .. package.cpath
    end
	if not package.path:find(cpath) then
		package.path = cpath .. "?.lua;" .. package.path
	end

	local expend_path = "/usr/local/share/lua/5.1/"
	if not package.path:find(expend_path) then
		package.path = expend_path .. "?.lua;" .. package.path
	end

	local db = nil
	local cache = nil
	
	local server_name,ip,today,day,body_length,method,config,cache_count
	local cache_count_threshold, cache_interval_threshold = 300, 10
	local cjson, sqlite3 
	local total_config, clients
	local update_day
	local data_dir
	local no_binding_server_name 
	local number_day
	local day_column
	local flow_column
	local spider_column

	local debug_mode = true
	
	local function debug(msg)
		if not debug_mode then return true end
    	local fp = io.open('/www/server/total/debug.log', 'ab')
    	if fp == nil then
        	return nil
    	end
		local localtime = os.date("%Y-%m-%d %H:%M:%S")
		if server_name then
    		-- fp:write(server_name.."/"..localtime..":"..tostring(msg) .. "\n")
    		fp:write(tostring(msg) .. "\n")
		else
    		fp:write(localtime..":"..tostring(msg) .. "\n")
		end
    	fp:flush()
    	fp:close()
    	return true
	end
	

	local function file_exists(file)
		local file = io.open(file, "rb")
		if file then 
			file:close()
			return true 
		end
		return false
	end
	
	local function check_dir(path)
		local file = io.open(path, "rb")
		if file then file:close() end
		return file ~= nil
	end
	
	local function is_migrating()
		local file = io.open("/www/server/total/migrating", "rb")
		if file then return true end
		return false
	end
	
	local function create_dir(path)
		if ngx.re.match(path,'/\\.\\./') then return false end
		local legal = true
		if not ngx.re.match(path,'^/www/server/total/') then 
			legal = false
		end
		if data_dir then
			if not ngx.re.match(path,'^'..data_dir) then
				legal = false
			else
				legal = true
			end
		end
		if not legal then
			return false
		end
		os.execute("mkdir -p " .. path)
	end
	
	local function write_file_bylog(filename,body,mode)
		local fp = io.open(filename,mode)
		if fp == nil then
			return nil
		end
		fp:write(body)
		fp:flush()
		fp:close()
		return true
	end
	
	local function read_file_body_bylog(filename)
		local fp = io.open(filename,'rb')
		if not fp then
			return nil
		end
		fbody = fp:read("*a")
		fp:close()
		if fbody == '' then
			return nil
		end
		return fbody
	end
	
	local function split_bylog(str,reps )
		local resultStrList = {}
		string.gsub(str,'[^'..reps..']+',function(w) table.insert(resultStrList,w) end)
		return resultStrList
	end

	local function arrlen_bylog(arr)
		if not arr then return 0 end
		count = 0
		for _,v in ipairs(arr) do
			count = count + 1
		end
		return count
	end
	
	local function is_ipaddr_bylog(client_ip)
		local cipn = split_bylog(client_ip,'.')
		if arrlen_bylog(cipn) < 4 then return false end
		for _,v in ipairs({1,2,3,4})
		do
			local ipv = tonumber(cipn[v])
			if ipv == nil then return false end
			if ipv > 255 or ipv < 0 then return false end
		end
		return true
	end
	
	local function get_client_ip_bylog()
		local client_ip = "unknown"
		local cdn = config['cdn']
		if cdn == true then
			for _,v in ipairs(config['cdn_headers']) do
				if request_header[v] ~= nil and request_header[v] ~= "" then
					local ip_list = request_header[v]
					client_ip = split_bylog(ip_list,',')[1]
					break;
				end
			end
		end
		if type(client_ip) == 'table' then client_ip = "" end
		if string.match(client_ip,"%d+%.%d+%.%d+%.%d+") == nil or not is_ipaddr_bylog(client_ip) then
			client_ip = ngx.var.remote_addr
			if client_ip == nil then
				client_ip = "unknown"
			end
		end
		return client_ip
	end

	
	local function arrip_bylog(ipstr)
		if ipstr == 'unknown' then return {0,0,0,0} end
		iparr = split_bylog(ipstr,'.')
		iparr[1] = tonumber(iparr[1])
		iparr[2] = tonumber(iparr[2])
		iparr[3] = tonumber(iparr[3])
		iparr[4] = tonumber(iparr[4])
		return iparr
	end
	
	local function get_request_time()
		local request_time = math.floor((ngx.now() - ngx.req.start_time()) * 1000)
		if request_time == 0 then  request_time = 1 end
		return request_time
	end
	
	local function get_length(k)
		local clen  = ngx.var.body_bytes_sent
		if clen == nil then clen = 0 end
		return tonumber(clen)
	end

	local function get_boundary()
		local header = request_header["content-type"]
		if not header then return nil end
		if type(header) == "table" then
			return false
		end
		if header then
			if ngx.re.find(header,[[multipart]],'ijo') then
				if not ngx.re.match(header,'^multipart/form-data; boundary=') then 
					return false
				end
			   	local multipart_data=ngx.re.match(header,'^multipart/form-data; boundary=.+')
				local check_file=ngx.re.gmatch(multipart_data[0],[[=]],'ijo')
				ret={}
				while true do
					local m, err = check_file()
					if m then 
						table.insert(ret,m)
					else
						break
					end 
				end
				if type(ret)~='table' then return false end 
				if(arrlen_bylog(ret)>=2) then
					return false
				end
				return header
			else
				return false
			end 
		end 
	end
	
	local function get_http_original()
		local data = ""
		local headers = ngx.req.get_headers()
		if not headers then return data end
		if method ~='GET' then 
			data = ngx.req.get_body_data()
			if not data then
				data = ngx.req.get_post_args(1000000)
			end
			if "string" == type(data) then
				headers["payload"] = data
			end

			if "table" == type(data) then
				headers = table.concat(headers, data)
			end
		end
		return json.encode(headers)
	end
	
	local function init_db(log_dir, server_name)
		-- 初始化日志db
		local path = log_dir .. "/" .. server_name
		if not check_dir(path) then 
			res = create_dir(path) 
			-- debug("create dir res:"..tostring(res))
		end
		local db_name = "logs.db"
		local db_path = path .. "/" .. db_name
		if ngx.shared.bt_total:get(db_path) then
			return
		end
		local db = sqlite3.open(db_path)
		local table_name = "site_logs"
		local stmt = db:prepare("SELECT COUNT(*) FROM sqlite_master where type='table' and name=?")
		local rows = 0
		if stmt ~= nil then
			stmt:bind_values(table_name)
			stmt:step()
			rows = stmt:get_uvalues()
			stmt:finalize()
		end
		if stmt ~= nil and rows >0 then
			ngx.shared.bt_total:set(db_path, true)
		else
			status, errorString = db:exec([[PRAGMA synchronous = 0]])
			status, errorString = db:exec([[PRAGMA page_size = 4096]])
			status, errorString = db:exec([[PRAGMA journal_mode = wal]])
			status, errorString = db:exec([[PRAGMA journal_size_limit = 1073741824]])
			status,errorString = db:exec[[
				CREATE TABLE site_logs (
					time INTEGER,
					ip TEXT,
					domain TEXT,
					server_name TEXT,
					method TEXT,
					status_code INTEGER,
					uri TEXT,
					body_length INTEGER,
					referer TEXT DEFAULT "",
					user_agent TEXT,
					is_spider INTEGER DEFAULT 0,
					protocol TEXT,
					request_time INTEGER,
					request_headers TEXT DEFAULT "",
					ip_list TEXT DEFAULT ""
				)]]
			status, errorString = db:exec([[CREATE INDEX time_inx ON site_logs(time)]])
			-- status, errorString = db:exec([[CREATE INDEX ip_inx ON site_logs(ip)]])
			-- status, errorString = db:exec([[CREATE INDEX uri_inx ON site_logs(uri)]])
			-- status, errorString = db:exec([[CREATE INDEX status_code_inx ON site_logs(status_code)]])
			-- status, errorString = db:exec([[CREATE INDEX method_inx ON site_logs(method)]])
			-- 客户端统计明细，按小时统计
			-- 客户端的定义: {client:"Android", column_name:"android"}
			status, errorString = db:exec([[
				CREATE TABLE client_stat(
					time INTEGER PRIMARY KEY,
					weixin INTEGER DEFAULT 0,
					android INTEGER DEFAULT 0,
					iphone INTEGER DEFAULT 0,
					mac INTEGER DEFAULT 0,
					windows INTEGER DEFAULT 0,
					linux INTEGER DEFAULT 0,
					edeg INTEGER DEFAULT 0,
					firefox INTEGER DEFAULT 0,
					msie INTEGER DEFAULT 0,
					metasr INTEGER DEFAULT 0,
					qh360 INTEGER DEFAULT 0,
					theworld INTEGER DEFAULT 0,
					tt INTEGER DEFAULT 0,
					maxthon INTEGER DEFAULT 0,
					opera INTEGER DEFAULT 0,
					qq INTEGER DEFAULT 0,
					uc INTEGER DEFAULT 0,
					pc2345 INTEGER DEFAULT 0, 
					safari INTEGER DEFAULT 0,
					chrome INTEGER DEFAULT 0,
					machine INTEGER DEFAULT 0,
					mobile INTEGER DEFAULT 0,
					other INTEGER DEFAULT 0,
					vivaldi INTEGER DEFAULT 0,
					brave INTEGER DEFAULT 0,
					chromium INTEGER DEFAULT 0
				)]])
			-- 请求分类统计明细，按小时统计
			-- time: 2021032500 时间刻度, 以小时为单位
			status, errorString = db:exec([[
				CREATE TABLE request_stat(
					time INTEGER PRIMARY KEY,
					req INTEGER DEFAULT 0,
					pv INTEGER DEFAULT 0,
					uv INTEGER DEFAULT 0,
					ip INTEGER DEFAULT 0,
					length INTEGER DEFAULT 0,
					spider INTEGER DEFAULT 0,
					status_500 INTEGER DEFAULT 0,
					status_501 INTEGER DEFAULT 0,
					status_502 INTEGER DEFAULT 0,
					status_503 INTEGER DEFAULT 0,
					status_504 INTEGER DEFAULT 0,
					status_505 INTEGER DEFAULT 0,
					status_506 INTEGER DEFAULT 0,
					status_507 INTEGER DEFAULT 0,
					status_509 INTEGER DEFAULT 0,
					status_510 INTEGER DEFAULT 0,
					status_400 INTEGER DEFAULT 0,
					status_401 INTEGER DEFAULT 0,
					status_402 INTEGER DEFAULT 0,
					status_403 INTEGER DEFAULT 0,
					status_404 INTEGER DEFAULT 0,
					http_get INTEGER DEFAULT 0,
					http_post INTEGER DEFAULT 0,
					http_put INTEGER DEFAULT 0,
					http_patch INTEGER DEFAULT 0,
					http_delete INTEGER DEFAULT 0
				)]])
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_405 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_406 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_407 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_408 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_409 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_410 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_411 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_412 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_413 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_414 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_415 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_416 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_417 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_418 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_421 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_422 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_423 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_424 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_425 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_426 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_449 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE request_stat ADD COLUMN status_451 INTEGER DEFAULT 0;")	

			-- 蜘蛛统计明细，按小时统计
			-- 按照已有蜘蛛分类，分别建立对应的统计列。
			status, errorString = db:exec[[
				CREATE TABLE spider_stat(
					time INTEGER PRIMARY KEY,
					google INTEGER DEFAULT 0,
					bing INTEGER DEFAULT 0,
					yahoo INTEGER DEFAULT 0,
					duckduckgo INTEGER DEFAULT 0,
					baidu INTEGER DEFAULT 0,
					yandex INTEGER DEFAULT 0,
					sogou INTEGER DEFAULT 0,
					exalead INTEGER DEFAULT 0,
					facebook INTEGER DEFAULT 0,
					alexa INTEGER DEFAULT 0,
					apple INTEGER DEFAULT 0,
					soso INTEGER DEFAULT 0,
					qh360 INTEGER DEFAULT 0,
					youdao INTEGER DEFAULT 0,
					dnspod INTEGER DEFAULT 0,
					yisou INTEGER DEFAULT 0,
					bytes INTEGER DEFAULT 0,
					other INTEGER DEFAULT 0
				)]]

			local sql_str1 = [[
				CREATE TABLE uri_stat (
					uri_md5 CHAR(32) PRIMARY KEY,
					uri TEXT,
					day1 INTEGER DEFAULT 0,
					day2 INTEGER DEFAULT 0,
					day3 INTEGER DEFAULT 0,
					day4 INTEGER DEFAULT 0,
					day5 INTEGER DEFAULT 0,
					day6 INTEGER DEFAULT 0,
					day7 INTEGER DEFAULT 0,
					day8 INTEGER DEFAULT 0,
					day9 INTEGER DEFAULT 0,
					day10 INTEGER DEFAULT 0,
					day11 INTEGER DEFAULT 0,
					day12 INTEGER DEFAULT 0,
					day13 INTEGER DEFAULT 0,
					day14 INTEGER DEFAULT 0,
					day15 INTEGER DEFAULT 0,
					day16 INTEGER DEFAULT 0,
					day17 INTEGER DEFAULT 0,
					day18 INTEGER DEFAULT 0,
					day19 INTEGER DEFAULT 0,
					day20 INTEGER DEFAULT 0,
					day21 INTEGER DEFAULT 0,
					day22 INTEGER DEFAULT 0,
					day23 INTEGER DEFAULT 0,
					day24 INTEGER DEFAULT 0,
					day25 INTEGER DEFAULT 0,
					day26 INTEGER DEFAULT 0,
					day27 INTEGER DEFAULT 0,
					day28 INTEGER DEFAULT 0,
					day29 INTEGER DEFAULT 0,
					day30 INTEGER DEFAULT 0,
					day31 INTEGER DEFAULT 0,
					day32 INTEGER DEFAULT 0
				)]]
			status, errorString = db:exec(sql_str1)	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow1 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow2 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow3 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow4 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow5 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow6 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow7 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow8 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow9 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow10 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow11 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow12 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow13 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow14 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow15 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow16 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow17 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow18 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow19 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow20 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow21 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow22 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow23 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow24 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow25 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow26 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow27 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow28 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow29 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow30 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE uri_stat ADD COLUMN flow31 INTEGER DEFAULT 0;")	

			debug("Init db status:"..tostring(status))
			debug("Init db error:"..tostring(errorString))

			local sql_str1 = [[
				CREATE TABLE ip_stat (
					ip CHAR(15) PRIMARY KEY,
					area CHAR(8) DEFAULT "",
					day1 INTEGER DEFAULT 0,
					day2 INTEGER DEFAULT 0,
					day3 INTEGER DEFAULT 0,
					day4 INTEGER DEFAULT 0,
					day5 INTEGER DEFAULT 0,
					day6 INTEGER DEFAULT 0,
					day7 INTEGER DEFAULT 0,
					day8 INTEGER DEFAULT 0,
					day9 INTEGER DEFAULT 0,
					day10 INTEGER DEFAULT 0,
					day11 INTEGER DEFAULT 0,
					day12 INTEGER DEFAULT 0,
					day13 INTEGER DEFAULT 0,
					day14 INTEGER DEFAULT 0,
					day15 INTEGER DEFAULT 0,
					day16 INTEGER DEFAULT 0,
					day17 INTEGER DEFAULT 0,
					day18 INTEGER DEFAULT 0,
					day19 INTEGER DEFAULT 0,
					day20 INTEGER DEFAULT 0,
					day21 INTEGER DEFAULT 0,
					day22 INTEGER DEFAULT 0,
					day23 INTEGER DEFAULT 0,
					day24 INTEGER DEFAULT 0,
					day25 INTEGER DEFAULT 0,
					day26 INTEGER DEFAULT 0,
					day27 INTEGER DEFAULT 0,
					day28 INTEGER DEFAULT 0,
					day29 INTEGER DEFAULT 0,
					day30 INTEGER DEFAULT 0,
					day31 INTEGER DEFAULT 0,
					day32 INTEGER DEFAULT 0
				)]]
			status,errorString = db:exec(sql_str1)
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow1 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow2 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow3 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow4 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow5 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow6 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow7 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow8 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow9 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow10 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow11 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow12 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow13 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow14 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow15 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow16 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow17 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow18 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow19 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow20 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow21 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow22 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow23 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow24 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow25 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow26 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow27 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow28 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow29 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow30 INTEGER DEFAULT 0;")	
			status, errorString = db:exec("ALTER TABLE ip_stat ADD COLUMN flow31 INTEGER DEFAULT 0;")	

			debug("Init ip_stat status:"..tostring(status))
			debug("Init ip_stat error:"..tostring(errorString))

			status, errorString = db:exec[[
				CREATE TABLE referer_stat(
					time INTEGER,
                	domain TEXT,
                	count INTEGER DEFAULT 0
				)
			]]

			ngx.shared.bt_total:set(db_path, true)
			ngx.log(ngx.INFO, "初始化日志数据库完成。")
			-- print(status, errorString)
		-- else
			-- ngx.log(ngx.INFO, "db 已存在。")
		end

		if db and db:isopen() then
			db:close()
		end
	end
	
	local function is_storing(name)
		local store_status = ngx.shared.bt_total:get(name.."_STORING")
		if store_status ~= nil and store_status == true then
			return true 
		end
		return false
	end
	
	local function get_end_time()
		local s_time = os.time()
		local n_date = os.date("*t",s_time + 86400)
		n_date.hour = 0
		n_date.min = 0
		n_date.sec = 0
		d_time = os.time(n_date)
		return d_time - s_time
	end
	
	local function get_data_on(filename)
		local data = ngx.shared.bt_total:get(filename)
		if not data then
			data = read_file_body_bylog(filename)
			if not data then
				data = {}
			else
				data = json.decode(data)
			end
		else
			data = json.decode(data)
		end
		return data
	end
	
	local function save_data_on(filename,data)
		local newdata = json.encode(data)
		local extime = 0;
		if string.find(filename,"%d+-%d+-%d+") then extime = 88400 end
		ngx.shared.bt_total:set(filename,newdata,extime)
		if not ngx.shared.bt_total:get(filename..'_lock') then
			ngx.shared.bt_total:set(filename..'_lock',1,5)
			write_file_bylog(filename,newdata,'wb')
		end
	end
	
	local function get_clients()
	-- local clients_json = cpath .. '/config/clients.json'
	-- local clients = get_data_on(clients_json)
		return require "clients"
	end
	
	local function match_pc_client(ua, clients)
		local msie_pattern = nil
		local safari_pattern = nil
		local chrome_pattern = nil
		for column, pattern in pairs(clients) do
			if column == "safari" then 
				safari_pattern = pattern
			elseif column == "chrome" then 
				chrome_pattern = pattern 
			elseif column == "msie" then 
				msie_pattern = pattern
			elseif ngx.re.match(ua, pattern, "jo") then
				return column
			end
		end
		if msie_pattern and ngx.re.match(ua, msie_pattern, "jo") then 
			return "msie"
		end
		if chrome_pattern and ngx.re.match(ua, chrome_pattern, "jo") then
			return "chrome"
		end
		if safari_pattern and ngx.re.match(ua, safari_pattern, "jo") then
			return "safari"
		end
		return nil
	end
	
	local function match_norepeat_client(ua, clients)
		if clients == nil then return end
		for column, pattern in pairs(clients) do
			if ngx.re.match(ua, pattern, "jo") then
				return column
			end
		end
	end

	local function get_update_field(field, value)
		return field.."="..field.."+"..tostring(value)
	end
	
	local function match_client()
		-- 匹配客户端
		local ua = ''
		if request_header['user-agent'] then
			ua = request_header['user-agent']
		end
		if not ua then
			return false, nil
		end
		local client_stat_fields = ""
		local clients_map = {
        	["android"] = "android",
        	["iphone"] = "iphone",
        	["ipod"] = "iphone",
        	["ipad"] = "iphone",
			["firefox"] = "firefox",
        	["chrome"] = "chrome",
        	["edge"] = "edeg",
        	["edg"] = "edeg",
        	["safari"] = "safari",
        	["msie"] = "msie",
        	["trident"] = "msie",
        	["opera"] = "opera",
        	["vivaldi"] = "vivaldi",
			["brave"] = "brave",
			["chromium"] = "chromium",
			["windows"] = "windows",
        	["linux"] = "linux",
        	["macintosh"] = "mac",
			["mobile"] = "mobile"
		}
		local mobile_regx = "(Mobile|Android|iPhone|iPod|iPad)"
		local mobile_res = ngx.re.match(ua, mobile_regx, "ijo")
		--mobile
		if mobile_res then
			client_stat_fields = client_stat_fields..","..get_update_field("mobile", 1)
			mobile_res = string.lower(mobile_res[0])
			if mobile_res ~= "mobile" then
				client_stat_fields = client_stat_fields..","..get_update_field(clients_map[mobile_res], 1)
			end
		else
			--pc
			-- 匹配结果的顺序，与ua中关键词的顺序有关
			-- lua的正则不支持|语法
			-- 短字符串string.find效率要比ngx正则高
			local pc_regx1 = "(Opera|Edg[e]*|Vivaldi|Brave|Chromium)" 
			local pc_res = ngx.re.match(ua, pc_regx1, "ijo")
			local cls_pc = nil
			if not pc_res then
				if string.find(ua, "[Ff]irefox") then
					cls_pc = "firefox"
				elseif string.find(ua, "MSIE") or string.find(ua, "Trident") then
					cls_pc = "msie"
				elseif string.find(ua, "[Cc]hrome") then
					cls_pc = "chrome"
				elseif string.find(ua, "[Ss]afari") then
					cls_pc = "safari"
				end
			else
				cls_pc = string.lower(pc_res[0])
			end
			-- debug("UA:"..ua)
			-- debug("PC cls:"..tostring(cls_pc))
			if cls_pc then
				client_stat_fields = client_stat_fields..","..get_update_field(clients_map[cls_pc], 1)
			else
				-- machine and other
				local machine_res, err = ngx.re.match(ua, "([Cc]url|HeadlessChrome|[a-zA-Z]+[Bb]ot|[Ww]get|[Ss]pider|[Cc]rawler|[Ss]crapy|zgrab|[Pp]ython|java)", "ijo")
				if machine_res then
					client_stat_fields = client_stat_fields..","..get_update_field("machine", 1)
				else
					-- 移动端+PC端+机器以外 归类到 其他
					client_stat_fields = client_stat_fields..","..get_update_field("other", 1)
				end
			end

			local os_regx = "(Windows|Linux|Macintosh)"
			local os_res = ngx.re.match(ua, os_regx, "ijo")
			if os_res then
				os_res = string.lower(os_res[0])
				client_stat_fields = client_stat_fields..","..get_update_field(clients_map[os_res], 1)
			end
		end

		-- local other_regx = "MicroMessenger"
		-- local other_res = string.find(ua, other_regx)
		-- if other_res then
		-- 	client_stat_fields = client_stat_fields..","..get_update_field("weixin", 1)
		-- end
		if client_stat_fields then
			client_stat_fields = string.sub(client_stat_fields, 2)
		end
		return client_stat_fields
	end
	
	local function get_spiders()
		local spiders = require "spiders"
		return spiders
	end
	
	local function match_spider()
		-- TODO 加入IP区间判断是否是蜘蛛请求
		-- 匹配蜘蛛请求
		local ua = ''
		if request_header['user-agent'] then
			ua = request_header['user-agent']
		end
		if not ua then
			return false, nil, 0
		end
		-- local spiders = get_spiders()
		local is_spider = false
		local spider_name = nil

		local spider_table = {
			["google"] = 1,
			["bing"] = 2,
			["yahoo"] = 3,
			["duckduckgo"] = 4,
			["baidu"] = 5,
			["yandex"] = 6,
			["sogou"] = 7,
			["exalead"] = 8,
			["facebook"] = 9,
			["alexa"] = 10,
			["apple"] = 11,
			["qh360"] = 12,
			["bytes"] = 13,
			["soso"] = 14,
			["dnspod"] = 15,
			["yisou"] = 16,
			["other"] = 17
		}

		local res,err = ngx.re.match(ua, "(Googlebot|AdsBot-Google|Google-Adwords|Bingbot|Slurp|Yahoo|DuckDuckBot|Baiduspider|Yandex|Sogou web spider|Exabot|facebot|ia_archiver|Applebot)", "ijo")
		if res then
			is_spider = true
			spider_match = string.lower(res[0])
			if string.find(spider_match, "google") then
				spider_name = "google"
			elseif string.find(spider_match, "bing") then
				spider_name = "bing"
			elseif string.find(spider_match, "yahoo") then
				spider_name = "yahoo"
			elseif string.find(spider_match, "slurp") then
				spider_name = "yahoo"
			elseif string.find(spider_match, "duckduckbot") then
				spider_name = "duckduckgo"
			elseif string.find(spider_match, "baidu") then
				spider_name = "baidu"
			elseif string.find(spider_match, "yandex") then
				spider_name = "yandex"
			elseif string.find(spider_match, "sogou") then
				spider_name = "sogou"
			elseif string.find(spider_match, "exabot") then
				spider_name = "exalead"
			elseif string.find(spider_match, "facebot") then
				spider_name = "facebook"
			elseif string.find(spider_match, "ia_archiver") then
				spider_name = "alexa"
			elseif string.find(spider_match, "applebot") then
				spider_name = "apple"
			end
		end

		if is_spider then 
			return is_spider, spider_name, spider_table[spider_name]
		end
		-- Curl|Yahoo|HeadlessChrome|包含bot|Wget|Spider|Crawler|Scrapy|zgrab|python|java|Adsbot|DuckDuckGo
		local other_res, err = ngx.re.match(ua, "(360Spider|Bytesspider|Sosospider|DNSPod-Monitor|YisouSpider)", "ijo")
		if other_res then
			-- debug("Other Spider:"..other_res[0].."/User Agent:"..tostring(ua))
			other_spider = string.lower(other_res[0])
			if string.find(other_spider, "360") then
				spider_name = "qh360"
			elseif string.find(other_spider, "bytes") then
				spider_name = "bytes"
			elseif string.find(other_spider, "soso") then
				spider_name = "soso"
			elseif string.find(other_spider, "dnspod") then
				spider_name = "dnspod"
			elseif string.find(other_spider, "yisou") then
				spider_name = "yisou"
			end
			if spider_name then
				return true, spider_name, spider_table[spider_name]
			end
		end
		return false, nil, 0
	end
	
	local function statistics_ipc()
		-- 判断IP是否重复的时间限定范围是请求的当前时间+24小时
		local ipc = 0
		local ip_token = server_name..'_'..ip
		if not ngx.shared.bt_total:get(ip_token) then
			ipc = 1
			ngx.shared.bt_total:set(ip_token,1,get_end_time())
		end
		return ipc
	end

	local function statistics_request()
		-- 计算pv uv
		local pvc = 0
		local uvc = 0

		if not is_spider and method == 'GET' and ngx.status == 200 and body_length > 512 then
			local ua = ''
			if request_header['user-agent'] then
				ua = string.lower(request_header['user-agent'])
			end
	
			out_header = ngx.resp.get_headers()
			if out_header['content-type'] then
				if string.find(out_header['content-type'],'text/html') then
					pvc = 1
					if request_header['user-agent'] then
						if string.find(ua,'mozilla') then
							local today = os.date("%Y-%m-%d")
							local uv_token = ngx.md5(ip .. request_header['user-agent'] .. today)
							if not ngx.shared.bt_total:get(uv_token) then
								uvc = 1
								ngx.shared.bt_total:set(uv_token,1,get_end_time())
							end
						end
					end
				end
			end
		end
		return pvc, uvc
	end
	
	local function total_flow()
		local flow_key = server_name .. '_flow'
		local flow_write = server_name .. '_flow_write'
		if ngx.shared.bt_total:get(flow_write) then
			ngx.shared.bt_total:incr(flow_key,body_length)
		else
			local write_time = ngx.now()
			local now_flow = ngx.shared.bt_total:get(flow_key)
			if not now_flow then now_flow = 0 end
			now_flow = now_flow + body_length
			local realtime_data = tostring(now_flow)..","..tostring(write_time)
	
			local queue_length = 10
			local position_key = flow_key.."_positon"
			local queue_key = flow_key.."_queue_"
	
			local req_position = ngx.shared.bt_total:incr(position_key, 1, 0)
			ngx.shared.bt_total:set(queue_key..tostring(req_position), realtime_data)
	
			local data = ""
			for i=1, queue_length, 1 do
				local _d = ngx.shared.bt_total:get(queue_key..tostring(i))
				if _d ~= nil then
					-- ngx.log(ngx.INFO, "------------------Realtime Req Queue:"..tostring(i).."/data:".._d.."----------------------------")
					data = data.."\n".._d 
				end
			end
	
			local path = cpath .. '/logs/' .. server_name
			if not check_dir(path) then create_dir(path) end
			local filename = path .. '/flow_sec.json'
			write_file_bylog(filename,tostring(data),'wb+')
	
			if req_position == queue_length then
				ngx.shared.bt_total:set(position_key, 0)
			end
	
			ngx.shared.bt_total:set(flow_key,body_length,2)
			ngx.shared.bt_total:set(flow_write,1,1)
		end
	end
	
	local function total_req_sec()
		local _key = server_name .. '_req'
		local _write = server_name .. '_req_write'
		if ngx.shared.bt_total:get(_write) then
			ngx.shared.bt_total:incr(_key,1)
		else
			local write_time = ngx.now()
			local now_ = ngx.shared.bt_total:get(_key)
			if not now_ then now_ = 0 end
			now_ = now_ + 1
			local realtime_data = tostring(now_)..","..tostring(write_time)
	
			local queue_length = 10
			local position_key = _key.."_positon"
			local queue_key = _key.."_queue_"
			local req_position = ngx.shared.bt_total:incr(position_key, 1, 0)
			-- ngx.log(ngx.INFO, "------------------Realtime Req Position:"..tostring(req_position).."----------------------------")
			ngx.shared.bt_total:set(queue_key..tostring(req_position), realtime_data)
			local data = ""
			for i=1, queue_length, 1 do
				local _d = ngx.shared.bt_total:get(queue_key..tostring(i))
				if _d ~= nil then
					-- ngx.log(ngx.INFO, "------------------Realtime Req Queue:"..tostring(i).."/data:".._d.."----------------------------")
					data = data.."\n".._d 
				end
			end
	
			local path = cpath .. 'logs/' .. server_name
			if not check_dir(path) then create_dir(path) end
			local filename = path .. '/req_sec.json'
			local write_res = write_file_bylog(filename,data,'wb+')
	
			if req_position == queue_length then
				ngx.shared.bt_total:set(position_key, 0)
			end
			ngx.shared.bt_total:set(_key,1,2)
			ngx.shared.bt_total:set(_write,1,1)
		end
	end
	
	local function error_log()
		if ngx.status ~= 500 and ngx.status ~= 502 and ngx.status ~= 401 and ngx.status ~= 503 then return false end
		local logline = {localtime , ip , server_name , method , ngx.status , ngx.var.request_uri , body_length , ngx.var.http_referer , request_header['user-agent'] , ngx.var.server_protocol}
		local path = cpath .. '/logs/' .. server_name .. '/error'
		if not check_dir(path) then create_dir(path) end
		write_file_bylog(path .. '/'.. tostring(ngx.status) .. '.log',json.encode(logline) .. "\n",'ab')
	end

	local function exclude_url()
		if not ngx.var.uri then return false end
		if not config['exclude_url'] then return false end
		local the_uri = string.sub(ngx.var.request_uri, 2)
		local url_conf = config["exclude_url"]
		for i,conf in pairs(url_conf)
		do
			local mode = conf["mode"]
			local url = conf["url"]
			if mode == "regular" then
				if ngx.re.find(the_uri, url, "ijo") then
					return true
				end
			else
				if the_uri == url then
					return true
				end
			end
		end
		return false
	end
	
	local function exclude_extension()
		if not ngx.var.uri then return false end
		if not config['exclude_extension'] then return false end
		for _,v in ipairs(config['exclude_extension'])
		do
			if ngx.re.find(ngx.var.uri,"[.]"..v.."$",'ijo') then
				return true
			end
		end
		return false
	end
	
	local function exclude_status()
		if not config['exclude_status'] then return false end
		local the_status = tostring(ngx.status)
		for _,v in ipairs(config['exclude_status'])
		do
			if the_status == v then
				return true
			end
		end
		return false
	end
	
	local function get_store_key()
		return os.date("%Y%m%d%H", os.time())
	end

	local function load_update_day(update_server_name)
		local _file = cpath.."logs/"..update_server_name.."/update_day.log"
		return read_file_body_bylog(_file)
	end

	local function write_update_day(update_server_name)
		local update_day = today
		local _file = cpath.."logs/"..update_server_name.."/update_day.log"
		write_file_bylog(_file, update_day, "w")
	end

	local function statistics_uri(db, stat_server_name, uri, uri_md5, body_length)
		-- 统计uri请求次数和流量
		local open_statistics_uri = config["statistics_uri"]
		if not open_statistics_uri then return true end

		local stat_sql = nil
		
		-- stat_sql = "INSERT INTO uri_stat(uri_md5,uri) SELECT \""..uri_md5.."\",\""..uri.."\" WHERE NOT EXISTS (SELECT uri_md5 FROM uri_stat WHERE uri_md5=\""..uri_md5.."\");"
		-- local res, err = db:exec(stat_sql)
		-- debug("Insert uri stat res:"..tostring(res))
		-- debug("Insert uri stat err:"..tostring(err))

		-- TODO 改预编译
		local stmt = db:prepare("INSERT INTO uri_stat(uri_md5, uri) SELECT :uri_md5, :uri WHERE NOT EXISTS (SELECT uri_md5 FROM uri_stat WHERE uri_md5 = :uri_md5);")
		if stmt then
			stmt:bind_names({uri_md5 = uri_md5, uri = uri})
			stmt:step()  
			stmt:reset()  
		end

		
		if is_spider then
			stat_sql = "UPDATE uri_stat SET "..day_column.."="..day_column.."+1,"..spider_column.."="..spider_column.."+"..body_length.." WHERE uri_md5=\""..uri_md5.."\""
		else
			stat_sql = "UPDATE uri_stat SET "..day_column.."="..day_column.."+1,"..flow_column.."="..flow_column.."+"..body_length.." WHERE uri_md5=\""..uri_md5.."\""
		end
		local res, err = db:exec(stat_sql)
		-- debug("Stat sql:"..stat_sql)
		-- debug("Update uri stat res:"..tostring(res))
		-- debug("Update uri stat err:"..tostring(err))
		return true
		-- local stmt = db:prepare(string.format("INSERT INTO %s(time) SELECT :time WHERE NOT EXISTS(SELECT time FROM %s WHERE time=:time);", stat_table, stat_table))
	end

	local function statistics_ip(db, stat_server_name, ip, body_length)
		local open_statistics_ip = config["statistics_ip"]	
		if not open_statistics_ip then return true end
		
		-- 原始语句
		-- local stat_sql = nil
		-- stat_sql = "INSERT INTO ip_stat(ip) SELECT \""..ip.."\" WHERE NOT EXISTS (SELECT ip FROM ip_stat WHERE ip=\""..ip.."\");"
		-- local res, err = db:exec(stat_sql)

		-- TODO 改预编译
		local stat_sql = nil
		stat_sql = nil
		local stmt, _ = db:prepare("INSERT INTO ip_stat(ip) SELECT :ip WHERE NOT EXISTS (SELECT ip FROM ip_stat WHERE ip=:ip);")
		if stmt then 
			stmt:bind_names {ip = ip}
			stmt:step()
			stmt:reset()
		end

		-- local hour_update_sql = "," .. day_hour1_column .. "=" .. day_hour1_column .. "+1," .. flow_hour1_column .. "=" .. flow_hour1_column .. "+" .. body_length
		-- stat_sql = "UPDATE ip_stat SET " .. day_column .. "=" .. day_column .. "+1," .. flow_column .. "=" .. flow_column .. "+" .. body_length .. hour_update_sql .. " WHERE ip=".. ngx.quote_sql_str(ip)
		-- local res, _ = db:exec(stat_sql)



		
		if is_spider then
			stat_sql = "UPDATE ip_stat SET "..day_column.."="..day_column.."+1,"..spider_column.."="..spider_column.."+"..body_length.." WHERE ip=\""..ip.."\""
		else
			stat_sql = "UPDATE ip_stat SET "..day_column.."="..day_column.."+1,"..flow_column.."="..flow_column.."+"..body_length.." WHERE ip=\""..ip.."\""
		end

		local res, err = db:exec(stat_sql)
		-- debug("Update ip res:"..tostring(res))
		-- debug("Update ip err:"..tostring(err))
		return true
	end
	
	local function update_stat(update_server_name, db, stat_table, key, columns)
		-- 根据指定表名，更新统计数据
		if not columns then return end
		
		local stmt = db:prepare(string.format("INSERT INTO %s(time) SELECT :time WHERE NOT EXISTS(SELECT time FROM %s WHERE time=:time);", stat_table, stat_table))
		stmt:bind_names{time=key}
		local res, err = stmt:step()
		stmt:finalize()
		local update_sql = "UPDATE ".. stat_table .. " SET " .. columns
		update_sql = update_sql .. " WHERE time=" .. key
		status, errorString = db:exec(update_sql)
	end



	local function get_server_name()
		local c_name = ngx.var.server_name
		-- debug("Ngx server name:"..c_name)
		local my_name = ngx.shared.bt_total:get(c_name)
		if my_name then return my_name end

		local domains = require "domains"
		for _,v in ipairs(domains)
		do
			if c_name == v["name"] then
				ngx.shared.bt_total:set(c_name,v['name'],3600)
				return v["name"]
			end
			for _,d_name in ipairs(v['domains'])
			do
				if c_name == d_name then
					ngx.shared.bt_total:set(c_name,v['name'],3600)
					return v['name']
	            elseif string.find(d_name, "*") then
					new_domain = string.gsub(d_name, '*', '.*')
					--debug("ngx server name:"..ngx.var.server_name.."/new_domain:"..new_domain)
	            	if string.find(c_name, new_domain) then
	                	ngx.shared.bt_total:set(c_name, v['name'],3600)
	                	return v['name']
	            	end
				end
			end
		end
		return no_binding_server_name
	end

	local function get_domain()
		local domain = request_header['host']
		if domain ~= nil then
			domain = string.gsub(domain, "_", ".")
		else
			domain = "unknown"
		end
		return domain
	end
		
	local function get_config(site, key)
		-- local config_data = json.decode(read_file_body_bylog(cpath..'/config.json'))
		local config_data = total_config
		local config = nil
		if config_data["global"] == nil then return nil end
		-- ngx.log(ngx.ERR, "------------------------CONFIG:"..json.encode(config_data).."----------------------")
		local global_config = config_data["global"]
		if key == "monitor" then
			if not global_config["monitor"] then
				return global_config["monitor"]
			end
		end
		if config_data[server_name] == nil then
			config = global_config
		else
			config = config_data[server_name]
			for k, v in pairs(global_config) do
				if config[k] == nil then
					config[k] = v
				end
			end
		end
		return config
	end

	local function cache_value(id, key, value)
		local line_id_key = "logline_"..server_name.."_"..id.."_"..key
		ngx.shared.bt_total:set(line_id_key, value)
	end

	local function clear_value(store_server_name, id, key)
		local line_id_key = "logline_"..store_server_name.."_"..id.."_"..key
		ngx.shared.bt_total:delete(line_id_key)
	end

	local function get_value(store_server_name, id, key)
		local line_id_key = "logline_"..store_server_name.."_"..id.."_"..key
		local value = ngx.shared.bt_total:get(line_id_key)
		-- ngx.shared.bt_total:delete(line_id_key)
		return value
	end

	-- 存储单行日志
	local function store_line(db, stmt, store_server_name, lineno)

		local logvalue = get_value(store_server_name, lineno, "logline")
		if not logvalue then return false end

		local logline = json.decode(logvalue)
		local time = logline["time"]
		local time_key = logline["time_key"]
		local ip = logline["ip"]
		local ip_list = logline["ip_list"]
		local domain = logline["domain"] 
		local server_name = logline["server_name"] 
		local real_server_name = logline["real_server_name"]
		local method = logline["method"] 
		local status_code = logline["status_code"] 
		local uri = logline["uri"] 
		local request_uri = logline["request_uri"] 
		local body_length = logline["body_length"] 
		local referer = logline["referer"] 
		local user_agent = logline["user_agent"] 
		local protocol = logline["protocol"] 
		local request_time = logline["request_time"] 
		local is_spider = logline["is_spider"] 
		local request_headers = logline["request_headers"] 
		local excluded = logline["excluded"] 
		local id = logline["id"]

		local request_stat_fields = nil 
		local client_stat_fields = nil
		local spider_stat_fields = nil
		local stat_fields = get_value(store_server_name, id, "STAT_FIELDS")
		if stat_fields == nil then
			debug("Log stat fields is nil.")
			debug("Logdata:"..logvalue)
		else
			stat_fields = split_bylog(stat_fields, ";")
			request_stat_fields = stat_fields[1]
			client_stat_fields = stat_fields[2]
			spider_stat_fields = stat_fields[3]

			if "x" == client_stat_fields then
				client_stat_fields = nil
			end

			if "x" == spider_stat_fields then
				spider_stat_fields = nil
			end
		end

		if not excluded then
			stmt:bind_names{
				time=time,
				ip=ip,
				domain=domain,
				server_name=real_server_name,
				method=method,
				status_code=status_code,
				uri=request_uri,
				body_length=body_length,
				referer=referer,
				user_agent=user_agent,
				protocol=protocol,
				request_time=request_time,
				is_spider=is_spider,
				request_headers=request_headers,
				ip_list=ip_list
			}
			local res, err = stmt:step()
			if tostring(res) == "5" then
				-- debug("Step res:"..tostring(res))
				-- debug("Step err:"..tostring(err))
				debug("Step数据库连接繁忙，稍候存储。")
				return false
			end
			stmt:reset()
			update_stat(store_server_name, db, "client_stat", time_key, client_stat_fields)
			update_stat(store_server_name, db, "spider_stat", time_key, spider_stat_fields)
			-- 只统计非蜘蛛请求
			local ok, err = pcall(function() statistics_uri(db, store_server_name, uri, ngx.md5(uri), body_length) end)
				-- debug("Statistics api result:"..tostring(ok))
				-- debug("Statistics api err:"..tostring(err))
			local ok, err = pcall(function() statistics_ip(db, store_server_name, ip, body_length) end)
				-- debug("Statistics ip result:"..tostring(ok))
				-- debug("Statistics ip err:"..tostring(err))
		end
		update_stat(store_server_name, db, "request_stat", time_key, request_stat_fields)
        return true
	end


	-- 存储指定站点的日志
	local function store_logs(store_server_name)
		if is_migrating() == true then
			debug("Migrating...")
			return
		end
		local flush_data = false
		local waiting_store_key = store_server_name .. "_WAITING_STORE"
		local flush_data_key = store_server_name .. "_FLUSH_DATA"
		local cache_count_id_key = store_server_name.."_CACHE_COUNT"
		local cache_count = ngx.shared.bt_total:get(cache_count_id_key)
		if cache_count == nil then cache_count = 0 end

        if not ngx.shared.bt_total:get(waiting_store_key) or cache_count > cache_count_threshold then
            local x = "go"
        else
            if ngx.shared.bt_total:get(flush_data_key) then
				flush_data = true
        	else
		       	-- debug("Cache count:"..cache_count)
                -- debug("Waiting...")
				return
			end
        end
		
		if is_storing(store_server_name) then
			debug("其他worker正在存储中，稍候存储。")
			ngx.shared.bt_total:delete(flush_data_key)
			return
		end
		local storing_key = store_server_name.."_STORING"
		local worker_id = ngx.worker.id()
		ngx.shared.bt_total:set(storing_key, true, 60)
		if flush_data == false then
			ngx.shared.bt_total:set(waiting_store_key, ngx.now(), cache_interval_threshold)
		end
		ngx.shared.bt_total:delete(flush_data_key)

		-- local store_start_time = gtime.get_msec()

		-- 1.计算存储数据区间段
		local last_insert_id_key = store_server_name.."_LAST_INSERT_ID"
		local store_start_id_key = store_server_name.."_STORE_START"
		local last_id = ngx.shared.bt_total:get(last_insert_id_key)
		local store_start = ngx.shared.bt_total:get(store_start_id_key)
		if store_start == nil then
			store_start = 1
		end
		local store_end = last_id
		if store_end == nil then
			store_end = 1
		end

		local site_config = get_config(store_server_name)

		-- ngx.log(ngx.ERR, "--------------["..worker_id.."]开始存储"..tostring(is_storing()).."/"..tostring(ngx.shared.bt_total:get("STORING")).."--------------------------")

		-- debug("--------------["..worker_id.."]开始存储"..tostring(is_storing()).."/"..tostring(ngx.shared.bt_total:get("STORING")).."--------------------------")
		-- debug("Start:"..store_start.."/End:"..store_end)
        local store_count = 0
		
		local stmt2 = nil
		
		local log_dir = cpath .. 'logs'
		data_dir = site_config["data_dir"]
		if data_dir then
			log_dir = data_dir
		end
		local db_path= log_dir .. '/' .. store_server_name .. "/logs.db"
		-- debug("db path:"..tostring(db_path))
		if not ngx.shared.bt_total:get(db_path) then
			debug("to init by shared dict control.")
			init_db(log_dir, store_server_name)
		end
		if not file_exists(db_path) then
			debug("to init by file check.")
			ngx.shared.bt_total:set(db_path, nil)
			init_db(log_dir, store_server_name)
		end
		local db, err = sqlite3.open(db_path)
		-- debug("Open db:"..tostring(db))
		-- debug("Open err:"..tostring(err))
		
		if db ~= nil then
			stmt2 = db:prepare[[INSERT INTO site_logs(
				time, ip, domain, server_name, method, status_code, uri, body_length,
				referer, user_agent, protocol, request_time, is_spider, request_headers, ip_list)
				VALUES(:time, :ip, :domain, :server_name, :method, :status_code, :uri,
				:body_length, :referer, :user_agent, :protocol, :request_time, :is_spider,
				:request_headers, :ip_list)]]
		end

		if db == nil or stmt2 == nil then
			debug("网站监控报表数据库连接异常。")
			ngx.shared.bt_total:set(storing_key, false)

			if db and db:isopen() then
				db:close()
			end
			return true
		end

		status, errorString = db:exec([[BEGIN TRANSACTION]])

		update_day = load_update_day(store_server_name)
		if not update_day or update_day ~= today then
			-- reset
			local update_sql = "UPDATE uri_stat SET "..day_column.."=0,"..spider_column.."=0,"..flow_column.."=0"
			status, errorString = db:exec(update_sql)

			update_sql = "UPDATE ip_stat SET "..day_column.."=0,"..flow_column.."=0"
			status, errorString = db:exec(update_sql)
			write_update_day(store_server_name)
		end
		-- local start_transaction_time = gtime.get_msec()
     	if store_end >= store_start then
			for i=store_start, store_end, 1 do
				if store_line(db, stmt2, store_server_name, i) then
					store_count = store_count + 1
					clear_value(store_server_name, i, "logline")
					clear_value(store_server_name, i, "STAT_FIELDS")
					-- if store_count >= cache_count_threshold then 
					-- 	store_end = i
					-- 	break
					-- end
				end
			end
		else
			local _tmp_store_end = store_end
			store_end = max_log_id
			for i=store_start, store_end, 1 do
				if store_line(db, stmt2, store_server_name, i) then
					store_count = store_count + 1
					clear_value(store_server_name, i, "logline")
					clear_value(store_server_name, i, "STAT_FIELDS")
				end
			end

			store_start = 1
			store_end = _tmp_store_end
			for i=store_start, store_end, 1 do
				if store_line(db, stmt2, store_server_name, i) then
					store_count = store_count + 1
					clear_value(store_server_name, i, "logline")
					clear_value(store_server_name, i, "STAT_FIELDS")
				end
			end
		end

		local res, err = stmt2:finalize()
		if tostring(res) == "5" then
			-- debug("Finalize res:"..tostring(res))
			-- debug("Finalize err:"..tostring(err))
			debug("数据库连接繁忙，稍候存储.")
		else

			-- local end_transaction_time = gtime.get_msec()
			-- debug("单行处理耗时："..tostring((end_transaction_time-start_transaction_time)/store_count))
			-- debug("事务处理耗时："..tostring(end_transaction_time-start_transaction_time))

			-- 清理过期日志
			-- local clear_start = gtime.get_msec()
			local now_date = os.date("*t")
			local _save_day = site_config["save_day"]
			local save_date_timestamp = os.time{year=now_date.year,
				month=now_date.month, day=now_date.day-_save_day, hour=0}
			local exec_res = db:exec("DELETE FROM site_logs WHERE time<"..tostring(save_date_timestamp))

			local res, err = db:execute([[COMMIT]])
			-- debug("Commit res:"..tostring(res))
			-- debug("Commit err:"..tostring(err))

		-- local end_clear = gtime.get_msec()
		-- debug("清理和提交事务耗时:"..tostring(end_clear-clear_start))
			ngx.shared.bt_total:set(store_start_id_key, store_end)
			ngx.shared.bt_total:incr(cache_count_id_key, tonumber(store_count)*-1)
			local cache_count = ngx.shared.bt_total:get(cache_count_id_key)
			if not cache_count then cache_count = 0 end
			if cache_count > cache_count_threshold or cache_count < 0 then
				ngx.shared.bt_total:delete(cache_count_id_key)
			end
		end

        if db and db:isopen() then
			db:close()
		end
		ngx.shared.bt_total:set(waiting_store_key, ngx.now(), cache_interval_threshold)
				-- local finish_time = gtime.get_msec()
		-- debug("-----------------------DEBUG: 存储总耗时:"..tostring(finish_time-store_start_time).."ms--------------------------")
        -- debug("Store count:"..tostring(store_count))
		ngx.shared.bt_total:set(storing_key, false)
	end


	-- 缓存并且存储日志
	local function cache_logs()

		-- local cache_start_time = gtime.get_msec()
		local excluded = false
		excluded = exclude_status() or exclude_extension() or exclude_url()

		request_header = ngx.req.get_headers()

		domain = get_domain()
		method = ngx.req.get_method()
		if method == "" or not method then return false end
		-- ngx.log(ngx.ERR, "Header:")
		-- for k,v in pairs(request_header) do
		-- 	ngx.log(ngx.ERR, "-----------------"..k..":"..v.."--------------------")
		-- end
		ip = get_client_ip_bylog()
		ip_list = request_header["x-forwarded-for"]
		if ip and not ip_list then
			ip_list = ip
		end
		-- ipn = arrip_bylog(ip)

		-- 缓存日志，只干塞缓存这一件事
		-- 记录从ngx启动以来生成的数据行id，进行累加
		local new_id = nil
		-- if ngx.shared.bt_total:get("STORE_START") then
		-- end
		local last_insert_id_key = server_name .. "_LAST_INSERT_ID"
		local cache_count_id_key = server_name .. "_CACHE_COUNT"
		if not ngx.shared.bt_total:get(server_name.."_CLEAR_CACHE_COUNT") then
			ngx.shared.bt_total:delete(server_name.."_CACHE_COUNT")
			ngx.shared.bt_total:set(server_name.."_CLEAR_CACHE_COUNT", true)
			-- debug("Cleared cache count.")
		end
		local err = nil
		new_id, err = ngx.shared.bt_total:incr(last_insert_id_key, 1, 0)
		ngx.shared.bt_total:incr(cache_count_id_key, 1, 0)
		if new_id >= max_log_id then
			ngx.shared.bt_total:set(last_insert_id_key, 1)
			new_id = ngx.shared.bt_total:get(last_insert_id_key)
		end

		-- debug("------------------["..server_name.."]STORE New ID:" ..tostring(new_id).."----------------------")
		local request_time = get_request_time()
		local status_code = ngx.status
		local referer = ngx.var.http_referer
		local uri = ngx.var.uri
		local request_uri = ngx.var.request_uri
		local protocol = ngx.var.server_protocol
		local time_key = get_store_key()
		local real_server_name = server_name
		if server_name == no_binding_server_name then
			real_server_name = ngx.var.server_name
		end
		
		local logline = {
			id=new_id,
			time_key=time_key,
			time=os.time(), 
			ip=ip, 
			domain=domain,
			server_name=server_name, 
			real_server_name=real_server_name,
			method=method, 
			status_code=status_code, 
			uri=uri, 
			request_uri=request_uri, 
			body_length=body_length, 
			referer=referer, 
			user_agent=request_header['user-agent'], 
			protocol=protocol,
			is_spider=0,
			request_time=request_time,
			excluded=excluded,
			request_headers="",
			ip_list=ip_list
		}

		local request_stat_fields = "req=req+1,length=length+"..body_length
		local spider_stat_fields = "x"
		local client_stat_fields = "x"
		local is_spider = false
		local request_spider = nil
		local spider_index = 0
		-- debug("Exclude:"..tostring(excluded).."/"..ngx.var.uri)
		if not excluded then

			if status_code == 500 or (method=="POST" and config["record_post_args"]==true) or (status_code==403 and config["record_get_403_args"]==true) then
				local data = ""
				local ok, err = pcall(function() data=get_http_original() end)
				if ok and not err then
					logline["request_headers"] = data
				end
				-- debug("Get http orgininal ok:"..tostring(ok))
				-- debug("Get http orgininal res:"..tostring(data))
				-- cache_value(new_id, "request_headers", ngx.req.raw_header())
			end

			if ngx.re.find("500,501,502,503,504,505,506,507,509,510,400,401,402,403,404,405,406,407,408,409,410,411,412,413,414,415,416,417,418,421,422,423,424,425,426,449,451", tostring(status_code), "jo") then
				local field = "status_"..status_code
				request_stat_fields = request_stat_fields .. ","..field.."="..field.."+1"
			end
			-- debug("method:"..method)
			local lower_method = string.lower(method)
			if ngx.re.find("get,post,put,patch,delete", lower_method, "ijo") then
				local field = "http_"..lower_method
				request_stat_fields = request_stat_fields .. ","..field.."="..field.."+1"
			end
     		local ipc = 0
     		local pvc = 0
     		local uvc = 0
			is_spider, request_spider, spider_index = match_spider()
			-- debug("Is spider:"..tostring(is_spider==true))
			-- debug("Request spider:".. tostring(request_spider))
			if not is_spider then
				client_stat_fields = match_client()	
				-- debug("Client stat fields:"..tostring(client_stat_fields))
				if not client_stat_fields or #client_stat_fields == 0 then
					client_stat_fields = request_stat_fields..",other=other+1"
				end

				if string.find(client_stat_fields, "machine") then
					if statistics_machine_access then
						pvc, uvc = statistics_request()
					end
					-- debug("no statistics.")
				else
					pvc, uvc = statistics_request()
				end
				ipc = statistics_ipc()
			else
				logline["is_spider"] = spider_index
				local field = "spider"
				request_stat_fields = request_stat_fields .. ","..field.."="..field.."+"..1
				spider_stat_fields = request_spider.."="..request_spider.."+"..1
			end

			if ipc > 0 then 
				request_stat_fields = request_stat_fields..",ip=ip+1"
			end
			if uvc > 0 then 
				request_stat_fields = request_stat_fields..",uv=uv+1"
			end
			if pvc > 0 then
				request_stat_fields = request_stat_fields..",pv=pv+1"
			end
		end
		
		local stat_fields = request_stat_fields..";"..client_stat_fields..";"..spider_stat_fields
		cache_value(new_id, "STAT_FIELDS", stat_fields)
		cache_value(new_id, "logline", json.encode(logline))
		-- debug("cached.")
	end

	local function run_logs()
		local presult, err = pcall(
			function() 
				json = require "cjson" 
				sqlite3 = require "lsqlite3"
				total_config = require "total_config"
			end
		)
		if not presult then
			debug("引入依赖错误："..tostring(err))
			return true
		end

		if ngx.var.uri == '/favicon.ico' or ngx.status == 0 or ngx.status == 444 then return true end
		no_binding_server_name = "btunknownbt"
		server_name = string.gsub(get_server_name(),'_','.')
		debug("server name:"..server_name)
		if server_name == 'default' or server_name == '_' or server_name == 'phpinfo' then return true end

		config = get_config(server_name)
		if config == nil then return true end
	
		monitor = config["monitor"]
		if monitor == false then
			-- debug("------------------------站点:"..server_name.."关闭监控----------------------")
			return true
		end
		max_log_id = 99999999999999
		body_length = get_length()

		today = os.date("%Y%m%d")

		day = os.date("%d")
		number_day = tonumber(day)
		day_column = "day"..number_day
		flow_column = "flow"..number_day
		spider_column = "spider_flow"..number_day

		if server_name == "127.0.0.1" then
			return true
		end
		
		total_flow()
		total_req_sec()

		statistics_machine_access = config["statistics_machine_access"]

		cache_logs()

		store_logs(server_name)
	end
	return run_logs()
}
